"""This  module defines the `BasePdf` that can be used to inherit from in order to build a custom PDF.

The `BasePDF` implements already a lot of ready-to-use functionality like integral, automatic normalization
and sampling.

Defining your own pdf
---------------------

A simple example:
>>> import zfit
>>> import zfit.z.numpy as znp
>>>
>>> class MyGauss(BasePDF):
>>>     def __init__(self, mean, stddev, name="MyGauss"):
>>>         super().__init__(mean=mean, stddev=stddev, name=name)
>>>
>>>     def _unnormalized_pdf(self, x):
>>>         return znp.exp((x - mean) ** 2 / (2 * stddev**2))

Notice that *here* we only specify the *function* and no normalization. This
**No** attempt to **explicitly** normalize the function should be done inside `_unnormalized_pdf`.
The normalization is handled with another method depending on the normalization range specified.
(It *is* possible, though discouraged, to directly provide the *normalized probability* by overriding _pdf(), but
there are other, more convenient ways to add improvements like providing an analytical integrals.)

Before we create an instance, we need to create the variables to initialize it
>>> mean = zfit.Parameter("mean1", 2., 0.1, 4.2)  # signature as in RooFit: *name, initial, lower, upper*
>>> stddev = zfit.Parameter("stddev1", 5., 0.3, 10.)
Let's create an instance and some example data
>>> gauss = MyGauss(mean=mean, stddev=stddev)
>>> example_data = np.random.random(10)
Now we can get the probability
>>> probs = gauss.pdf(x=example_data,norm_range=(-30., 30))  # `norm_range` specifies over which range to normalize
Or the integral
>>> integral = gauss.integrate(limits=(-5, 3.1), norm_range=False)  # norm_range is False -> return unnormalized
integral
Or directly sample from it
>>> sample = gauss.sample(n_draws=1000, limits=(-10, 10))  # draw 1000 samples within (-10, 10)

We can create an extended PDF, which will result in anything using a `norm_range` to not return the
probability but the number probability (the function will be normalized to `yield` instead of 1 inside
the `norm_range`)
>>> yield1 = Parameter("yield1", 100, 0, 1000)
>>> gauss_extended = gauss.create_extended(yield1)
>>> gauss.is_extended
True

>>> integral_extended = gauss.integrate(limits=(-10, 10), norm_range=(-10, 10))  # yields approx 100

For more advanced methods and ways to register analytic integrals or overwrite certain methods, see
also the advanced models in `zfit models <https://github.com/zfit/zfit-tutorials>`_
"""

#  Copyright (c) 2021 zfit

import warnings
from contextlib import suppress
from typing import Dict, Optional, Set, Type, Union

import tensorflow as tf

import zfit.z.numpy as znp
from zfit import z

from ..settings import run, ztypes
from ..util import ztyping
from ..util.cache import invalidate_graph
from ..util.deprecation import deprecated
from ..util.exception import (AlreadyExtendedPDFError, BreakingAPIChangeError,
                              FunctionNotImplemented, NotExtendedPDFError,
                              SpecificFunctionNotImplemented)
from ..util.temporary import TemporarilySet
from .basemodel import BaseModel
from .baseobject import extract_filter_params
from .interfaces import ZfitParameter, ZfitPDF
from .parameter import Parameter, convert_to_parameter
from .sample import extended_sampling
from .space import Space

_BasePDF_USER_IMPL_METHODS_TO_CHECK = {}


def _BasePDF_register_check_support(has_support: bool):
    """Marks a method that the subclass either *has* to or *can't* use the `@supports` decorator.

    Args:
        has_support: If True, flags that it **requires** the `@supports` decorator. If False,
            flags that the `@supports` decorator is **not allowed**.
    """
    if not isinstance(has_support, bool):
        raise TypeError("Has to be boolean.")

    def register(func):
        """Register a method to be checked to (if True) *has* `support` or (if False) has *no* `support`.

        Args:
            func:

        Returns:
            Function:
        """
        name = func.__name__
        _BasePDF_USER_IMPL_METHODS_TO_CHECK[name] = has_support
        func.__wrapped__ = _BasePDF_register_check_support
        return func

    return register


class BasePDF(ZfitPDF, BaseModel):

    def __init__(self, obs: ztyping.ObsTypeInput, params: Dict[str, ZfitParameter] = None, dtype: Type = ztypes.float,
                 name: str = "BasePDF",
                 **kwargs):
        super().__init__(obs=obs, dtype=dtype, name=name, params=params, **kwargs)

        self._yield = None
        self._norm_range = None
        self._normalization_value = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._subclass_check_support(methods_to_check=_BasePDF_USER_IMPL_METHODS_TO_CHECK,
                                    wrapper_not_overwritten=_BasePDF_register_check_support)

    # @property
    # def space(self) -> "zfit.Space":
    #     if self._norm_range is not None:
    #         space = self._norm_range
    #     else:
    #         space = super().space
    #
    #     return space

    def _check_input_norm_range(self, norm_range, none_is_error=False):
        if norm_range is None:
            norm_range = self.norm_range
        return super()._check_input_norm_range(norm_range=norm_range, none_is_error=none_is_error)

    def _check_input_params(self, *params):
        return tuple(convert_to_parameter(p) for p in params)

    def _func_to_integrate(self, x: ztyping.XType):
        return self.pdf(x, norm_range=False)

    def _func_to_sample_from(self, x):
        return self.pdf(x, norm_range=False)

    @property
    def norm_range(self) -> Union[Space, None, bool]:
        """Return the current normalization range. If None and the `obs` have limits, they are returned.

        Returns:
            The current normalization range.
        """
        norm_range = self._norm_range
        if norm_range is None:
            norm_range = self.space
        return norm_range

    @invalidate_graph
    def set_norm_range(self, norm_range: ztyping.LimitsTypeInput):
        """Set the normalization range (temporarily if used with contextmanager).

        Args:
            norm_range:
        """
        norm_range = self._check_input_norm_range(norm_range=norm_range)

        def setter(value):
            self._norm_range = value

        def getter():
            return self._norm_range

        return TemporarilySet(value=norm_range, setter=setter, getter=getter)

    # TODO: remove below?
    # @property
    # def _yield(self):
    #     """For internal use, the yield or None"""
    #     return self.params.get('yield')
    #
    # @_yield.setter
    # def _yield(self, value):
    #     if value is None:
    #         # unset
    #         self._params.pop('yield', None)  # safely remove if still there
    #     else:
    #         self._params['yield'] = value

    @_BasePDF_register_check_support(True)
    def _normalization(self, limits):
        raise SpecificFunctionNotImplemented

    def normalization(self, limits: ztyping.LimitsType) -> ztyping.XType:
        """Return the normalization of the function (usually the integral over `limits`).

        Args:
            limits:  The limits on where to normalize over

        Returns:
            The normalization value
        """
        limits = self._check_input_limits(limits=limits)

        return self._single_hook_normalization(limits=limits)

    def _single_hook_normalization(self, limits):  # TODO(Mayou36): add yield?
        return self._hook_normalization(limits=limits)

    def _hook_normalization(self, limits):
        return self._call_normalization(limits=limits)  # no _norm_* needed

    def _call_normalization(self, limits):
        # TODO: caching? alternative
        with suppress(FunctionNotImplemented):
            return self._normalization(limits=limits)
        return self._fallback_normalization(limits)

    def _fallback_normalization(self, limits):
        return self._hook_integrate(limits=limits, norm_range=False)

    def _unnormalized_pdf(self, x):
        raise SpecificFunctionNotImplemented

    @deprecated(None, "Use `pdf(norm_range=False)` instead")
    def unnormalized_pdf(self, x: ztyping.XType, component_norm_range: ztyping.LimitsTypeInput = None) -> ztyping.XType:
        """PDF "unnormalized". Use `functions` for unnormalized pdfs. this is only for performance in special cases.

        Args:
            x: The value, have to be convertible to a Tensor
            component_norm_range: The normalization range for the components. Needed for
            certain composition
                pdfs.

        Returns:
            1-dimensional :py:class:`tf.Tensor` containing the unnormalized pdf.
        """
        if component_norm_range is not None:
            raise BreakingAPIChangeError("component norm range should not be given anymore. If you want to set the norm"
                                         " range for the components, use `set_norm_range(..., propagate=True)")
        with self._convert_sort_x(x) as x:
            return self._single_hook_unnormalized_pdf(x)

    def _single_hook_unnormalized_pdf(self, x):
        return self._call_unnormalized_pdf(x=x)

    def _call_unnormalized_pdf(self, x):
        # try:
        return self._unnormalized_pdf(x)

    # except ValueError as error:
    #     raise ShapeIncompatibleError("Most probably, the number of obs the pdf was designed for"
    #                                  "does not coincide with the `n_obs` from the `space`/`obs`"
    #                                  "it received on initialization."
    #                                  "Original Error: {}".format(error))

    @z.function(wraps='model')
    def ext_pdf(self, x: ztyping.XTypeInput, norm_range: ztyping.LimitsTypeInput = None) -> ztyping.XType:
        """Probability density function scaled by yield, normalized over `norm_range`.

        Args:
          x: `float` or `double` `Tensor`.
          norm_range: :py:class:`~zfit.Space` to normalize over

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        if not self.is_extended:
            raise NotExtendedPDFError(f"{self} is not extended, cannot call `ext_pdf`")
        return self.pdf(x=x, norm_range=norm_range) * self.get_yield()

    @z.function(wraps='model')
    def ext_log_pdf(self, x: ztyping.XTypeInput, norm_range: ztyping.LimitsTypeInput = None) -> ztyping.XType:
        """Log of probability density function scaled by yield, normalized over `norm_range`.

        Args:
          x: `float` or `double` `Tensor`.
          norm_range: :py:class:`~zfit.Space` to normalize over

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        if not self.is_extended:
            raise NotExtendedPDFError(f"{self} is not extended, cannot call `ext_pdf`")
        return self.log_pdf(x=x, norm_range=norm_range) + znp.log(self.get_yield())

    @_BasePDF_register_check_support(False)
    def _pdf(self, x, norm_range):
        raise SpecificFunctionNotImplemented

    @z.function(wraps='model')
    def pdf(self, x: ztyping.XTypeInput, norm_range: ztyping.LimitsTypeInput = None) -> ztyping.XType:
        """Probability density function, normalized over `norm_range`.

        Args:
          x: `float` or `double` `Tensor`.
          norm_range: :py:class:`~zfit.Space` to normalize over

        Returns:
          :py:class:`tf.Tensor` of type `self.dtype`.
        """
        norm_range = self._check_input_norm_range(norm_range, none_is_error=True)
        with self._convert_sort_x(x) as x:
            value = self._single_hook_pdf(x=x, norm_range=norm_range)
            if run.numeric_checks:
                z.check_numerics(value, message="Check if pdf output contains any NaNs of Infs")
            return z.to_real(value)

    def _single_hook_pdf(self, x, norm_range):
        return self._hook_pdf(x=x, norm_range=norm_range)

    def _hook_pdf(self, x, norm_range):
        return self._norm_pdf(x=x, norm_range=norm_range)

    def _norm_pdf(self, x, norm_range):
        return self._call_pdf(x=x, norm_range=norm_range)

    def _call_pdf(self, x, norm_range):
        with suppress(FunctionNotImplemented):
            return self._pdf(x, norm_range=norm_range)
        with suppress(FunctionNotImplemented):
            return znp.exp(self._log_pdf(x=x, norm_range=norm_range))
        return self._fallback_pdf(x=x, norm_range=norm_range)

    def _fallback_pdf(self, x, norm_range):
        pdf = self._call_unnormalized_pdf(x)
        if norm_range.has_limits:
            pdf /= self._hook_normalization(limits=norm_range)
        return pdf

    @_BasePDF_register_check_support(False)
    def _log_pdf(self, x, norm_range):
        raise SpecificFunctionNotImplemented

    def log_pdf(self, x: ztyping.XType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
        """Log probability density function normalized over `norm_range`.

        Args:
          x: `float` or `double` `Tensor`.
          norm_range: :py:class:`~zfit.Space` to normalize over

        Returns:
          A `Tensor` of type `self.dtype`.
        """
        norm_range = self._check_input_norm_range(norm_range)
        with self._convert_sort_x(x) as x:
            return self._single_hook_log_pdf(x=x, norm_range=norm_range)

    def _single_hook_log_pdf(self, x, norm_range):
        return self._hook_log_pdf(x=x, norm_range=norm_range)

    def _hook_log_pdf(self, x, norm_range):
        log_prob = self._norm_log_pdf(x=x, norm_range=norm_range)
        return log_prob

    def _norm_log_pdf(self, x, norm_range):
        return self._call_log_pdf(x=x, norm_range=norm_range)

    def _call_log_pdf(self, x, norm_range):
        with suppress(FunctionNotImplemented):
            return self._log_pdf(x=x, norm_range=norm_range)
        with suppress(FunctionNotImplemented):
            return znp.log(self._pdf(x=x, norm_range=norm_range))
        return self._fallback_log_pdf(x=x, norm_range=norm_range)

    def _fallback_log_pdf(self, x, norm_range):
        return znp.log(self._hook_pdf(x=x, norm_range=norm_range))

    def gradient(self, x: ztyping.XType, norm_range: ztyping.LimitsType, params: ztyping.ParamsTypeOpt = None):
        raise BreakingAPIChangeError("Removed with 0.5.x: is this needed?")

    @z.function(wraps='model')
    def ext_integrate(self, limits: ztyping.LimitsType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
        """Integrate the function over `limits` (normalized over `norm_range` if not False).

        Args:
            limits: the limits to integrate over
            norm_range: the limits to normalize over or False to integrate the
                unnormalized probability

        Returns:
            The integral value as a scalar with shape ()
        """
        norm_range = self._check_input_norm_range(norm_range)
        limits = self._check_input_limits(limits=limits)
        if not self.is_extended:
            raise NotExtendedPDFError(f"{self} is not extended, cannot call `ext_pdf`")
        return self.integrate(limits=limits, norm_range=norm_range) * self.get_yield()

    def _apply_yield(self, value: float, norm_range: ztyping.LimitsType, log: bool) -> Union[float, tf.Tensor]:
        if self.is_extended and not norm_range.limits_are_false:
            if log:
                value += znp.log(self.get_yield())
            else:
                value *= self.get_yield()
        return value

    def apply_yield(self, value: Union[float, tf.Tensor], norm_range: ztyping.LimitsTypeInput = False,
                    log: bool = False) -> Union[float, tf.Tensor]:
        """If a norm_range is given, the value will be multiplied by the yield.

        Args:
            value:
            norm_range:
            log:

        Returns:
            Numerical
        """
        norm_range = self._check_input_norm_range(norm_range=norm_range)
        return self._apply_yield(value=value, norm_range=norm_range, log=log)

    @deprecated(None, "Use the public `set_yield` instead.")
    def _set_yield_inplace(self, value: Union[ZfitParameter, float, None]):
        """Make the model extended by setting a yield.

        This does not alter the general behavior of the PDF. If there is a
        `norm_range` given, the output of the above functions does not represent a normalized
        probability density function anymore but corresponds to a number probability.

        Args:
            value:
        """

        self._set_yield(value=value)

    def create_extended(self, yield_: ztyping.ParamTypeInput, name_addition="_extended") -> "ZfitPDF":
        """Return an extended version of this pdf with yield `yield_`. The parameters are shared.

        Args:
            yield_:
            name_addition:

        Returns:
            :py:class:`~zfit.core.interfaces.ZfitPDF`
        """
        # TODO(Mayou36): fix copy
        from zfit.models.functor import ProductPDF
        if isinstance(self, ProductPDF):
            warnings.warn(
                "As `copy` is not yet properly implemented, this may fails (for ProductPDF for example?). This"
                "will be fixed in the future.")
        if self.is_extended:
            raise AlreadyExtendedPDFError("This PDF is already extended, cannot create an extended one.")
        new_pdf = self.copy(name=self.name + str(name_addition))
        new_pdf.set_yield(value=yield_)
        return new_pdf

    def set_yield(self, value):

        """Make the model extended by setting a yield. If possible, prefer to use `create_extended`.

        This does not alter the general behavior of the PDF. The `pdf` and `integrate` and similar methods will
        continue to return the same - normalized to 1 - values. However, not only can this parameter be accessed
        via `get_yield`, the methods `ext_pdf` and `ext_integral` provide a version of `pdf` and `integrate`
        respecetively that is multiplied by the yield.

        These can be useful for plotting and for binned likelihoods.

        Args:
            value ():
        """
        self._set_yield(value=value)

    def _set_yield(self, value: ztyping.ParamTypeInput):
        if value is None:
            raise BreakingAPIChangeError("Cannot unset a yield (anymore).")
        if self.is_extended:
            raise AlreadyExtendedPDFError(f"Cannot extend {self}, is already extended.")
        value = convert_to_parameter(value)
        self.add_cache_deps(value)
        self._yield = value

    @property
    def is_extended(self) -> bool:
        """Flag to tell whether the model is extended or not.

        Returns:
            A boolean.
        """
        return self._yield is not None

    def _hook_sample(self, limits, n):
        if n is None and self.is_extended:
            n = 'extended'
        if isinstance(n, str) and n == 'extended':
            if not self.is_extended:
                raise NotExtendedPDFError("Cannot use 'extended' as value for `n` on a non-extended pdf.")
            samples = extended_sampling(pdfs=self, limits=limits)
        elif isinstance(n, str):
            raise ValueError("`n` is a string and not 'extended'. Other options are currently not implemented.")
        elif n is None:
            raise tf.errors.InvalidArgumentError("`n` cannot be `None` if pdf is not extended.")
        else:
            samples = super()._hook_sample(limits=limits, n=n)
        return samples

    def get_yield(self) -> Union[Parameter, None]:
        """Return the yield (only for extended models).

        Returns:
            The yield of the current model or None
        """
        # if not self.is_extended:
        #     raise zexception.ExtendedPDFError("PDF is not extended, cannot get yield.")
        return self._yield

    def _get_params(self,
                    floating: Optional[bool] = True,
                    is_yield: Optional[bool] = None,
                    extract_independent: Optional[bool] = True) -> Set[ZfitParameter]:

        params = super()._get_params(floating, is_yield=is_yield,
                                     extract_independent=extract_independent)

        if is_yield is not False:
            if self.is_extended:
                yield_params = extract_filter_params(self.get_yield(), floating=floating,
                                                     extract_independent=extract_independent)
                yield_params.update(params)  # putting the yields at the beginning
                params = yield_params
            elif is_yield is True:
                raise NotExtendedPDFError("PDF is not extended but only yield parameters were requested.")
        return params

    def create_projection_pdf(self, limits_to_integrate: ztyping.LimitsTypeInput) -> 'ZfitPDF':
        """Create a PDF projection by integrating out some of the dimensions.

        The new projection pdf is still fully dependent on the pdf it was created with.

        Args:
            limits_to_integrate:

        Returns:
            A pdf without the dimensions from `limits_to_integrate`.
        """
        from ..models.special import SimpleFunctorPDF

        def partial_integrate_wrapped(self_simple, x):
            return self.partial_integrate(x, limits=limits_to_integrate, norm_range=False)

        new_pdf = SimpleFunctorPDF(obs=self.space.get_subspace(obs=[obs for obs in self.obs
                                                                    if obs not in limits_to_integrate.obs]),
                                   pdfs=(self,),
                                   func=partial_integrate_wrapped)
        return new_pdf

    def copy(self, **override_parameters) -> 'BasePDF':
        """Creates a copy of the model.

        Note: the copy model may continue to depend on the original
        initialization arguments.

        Args:
          **override_parameters: String/value dictionary of initialization
            arguments to override with new value.

        Returns:
          A new instance of `type(self)` initialized from the union
            of self.parameters and override_parameters, i.e.,
            `dict(self.parameters, **override_parameters)`.
        """
        obs = self.norm_range

        # HACK(Mayou36): remove once copy is proper implemented
        from ..models.dist_tfp import WrapDistribution
        from ..models.kde import GaussianKDE1DimV1
        from ..models.polynomials import RecursivePolynomial

        if type(self) == WrapDistribution:  # NOT isinstance! Because e.g. Gauss wraps that and takes different args
            parameters = dict(distribution=self._distribution, dist_params=self.dist_params)
        else:
            # HACK END

            parameters = dict(self.params)
            lam = parameters.pop('lambda', None)
            if lam is not None:
                parameters['lam'] = lam

        if type(self) == GaussianKDE1DimV1:
            raise RuntimeError("Cannot copy `GaussianKDE1DimV1` (yet). If you tried to make it extended, use "
                               "`set_yield`"
                               " instead and set it inplace.")
            parameters['data'] = self._original_data

        # HACK(Mayou36): copy the polynomial correct, replace 'c_0' with coeff0/coeff_0 or similar
        if isinstance(self, RecursivePolynomial):
            parameters['coeff0'] = parameters.pop('c_0', None)
            coeffs = []
            i_coeff = 1
            # collect coeffs and convert to 'coeff' list
            while True:
                coeff_name = f'c_{i_coeff}'
                try:
                    coeff = parameters.pop(coeff_name)
                except KeyError:
                    break
                else:
                    coeffs.append(coeff)
                i_coeff += 1
            parameters['coeffs'] = coeffs

        from zfit.models.functor import BaseFunctor, SumPDF
        if isinstance(self, BaseFunctor):
            parameters = {}
            if isinstance(self, SumPDF):
                fracs = self.fracs
                if not self.is_extended:
                    fracs = fracs[:-1]
                parameters.update(fracs=fracs)
            parameters.update(pdfs=self.pdfs)

        parameters.update(obs=obs, name=self.name)
        parameters.update(**override_parameters)
        # if hasattr(self, "distribution"):
        #     parameters.update(distribution=self.distribution)
        yield_ = parameters.pop('yield', None)
        new_instance = type(self)(**parameters)
        if yield_ is not None:
            new_instance.set_yield(yield_)
        return new_instance

    def as_func(self, norm_range: ztyping.LimitsType = False):
        """Return a `Function` with the function `model(x, norm_range=norm_range)`.

        Args:
            norm_range:
        """
        from .operations import convert_pdf_to_func  # prevent circular import

        return convert_pdf_to_func(pdf=self, norm_range=norm_range)

    def __str__(self):
        return ("zfit.model.{type_name}("
                "\"{self_name}\""
                ", dtype={dtype})".format(
            type_name=type(self).__name__,
            self_name=self.name,
            dtype=self.dtype.name))



#  Copyright (c) 2021 zfit
import pytest

import zfit


def test_get_dependents_is_deterministic():
    parameters = [zfit.Parameter(f'param{i}', i) for i in range(4)]
    obs = zfit.Space('obs1', (-3, 2))

    def create_pdf(params):
        param1, param2, param3, param4 = params
        gauss1 = zfit.pdf.Gauss(param1, param2, obs=obs)
        gauss2 = zfit.pdf.Gauss(param2, param2, obs=obs)
        gauss3 = zfit.pdf.Gauss(param2, param4, obs=obs)
        gauss4 = zfit.pdf.Gauss(param3, param1, obs=obs)
        sum_pdf = zfit.pdf.SumPDF([gauss1, gauss2, gauss3, gauss4], fracs=[0.1, 0.3, 0.4])
        return sum_pdf

    for params in (parameters, reversed(parameters)):
        pdf = create_pdf(params)
        assert pdf.get_cache_deps() == pdf.get_cache_deps(), "get_dependents() not deterministic"


def test_get_params():
    obs = zfit.Space('obs', (-4, 5))
    mu_nofloat = zfit.Parameter('mu_nofloat', 1, floating=False)
    mu2 = zfit.Parameter('mu2', 1)
    sigma2 = zfit.Parameter('sigma2', 2)
    sigma_comp = zfit.ComposedParameter('sigma_comp', lambda s: s * 0.7, params=sigma2)

    yield1 = zfit.Parameter('yield1', 10)
    yield2 = zfit.Parameter('yield2', 200)
    yield2_comp = zfit.ComposedParameter('yield2_comp', lambda y: y * 0.9, params=yield2)

    gauss = zfit.pdf.Gauss(mu_nofloat, sigma_comp, obs)
    gauss2 = zfit.pdf.Gauss(mu2, sigma2, obs)
    gauss_ext = gauss.create_extended(yield1)
    gauss2_ext = gauss2.create_extended(yield2_comp)

    frac = zfit.Parameter("frac", 0.4)
    sum1 = zfit.pdf.SumPDF([gauss, gauss2], fracs=frac)
    sum1_ext = zfit.pdf.SumPDF([gauss_ext, gauss2_ext])

    assert set(gauss.get_params()) == {sigma2}

    with pytest.raises(ValueError):
        set(gauss.get_params(floating=True, is_yield=False, extract_independent=False))

    with pytest.raises(ValueError):
        set(gauss_ext.get_params(floating=True, is_yield=False, extract_independent=False))
    assert set(gauss_ext.get_params(floating=None, is_yield=False, extract_independent=False)) == {mu_nofloat,
                                                                                                   sigma_comp}
    with pytest.raises(ValueError):
        set(gauss_ext.get_params(floating=False, is_yield=False, extract_independent=False))

    with pytest.raises(ValueError):
        set(gauss_ext.get_params(floating=True, is_yield=None, extract_independent=False))

    assert set(gauss_ext.get_params(floating=None, is_yield=None, extract_independent=False)) == {mu_nofloat,
                                                                                                  sigma_comp, yield1}
    with pytest.raises(ValueError):
        set(gauss_ext.get_params(floating=False, is_yield=None, extract_independent=False))

    assert set(gauss_ext.get_params(floating=False, is_yield=True, extract_independent=False)) == set()
    assert set(gauss_ext.get_params(floating=None, is_yield=True, extract_independent=False)) == {yield1}
    assert set(gauss_ext.get_params(floating=True, is_yield=True, extract_independent=False)) == {yield1}

    # with extract deps
    assert set(gauss_ext.get_params(floating=True, is_yield=False, extract_independent=True)) == {sigma2}
    assert set(gauss_ext.get_params(floating=None, is_yield=False,
                                    extract_independent=True)) == {mu_nofloat, sigma2}
    assert set(gauss_ext.get_params(floating=False, is_yield=False,
                                    extract_independent=True)) == {mu_nofloat}

    assert set(gauss_ext.get_params(floating=True, is_yield=None,
                                    extract_independent=True)) == {sigma2, yield1}
    assert set(gauss_ext.get_params(floating=None, is_yield=None,
                                    extract_independent=True)) == {mu_nofloat, sigma2, yield1}
    assert set(gauss_ext.get_params(floating=False, is_yield=None,
                                    extract_independent=True)) == {mu_nofloat}

    assert set(gauss_ext.get_params(floating=True, is_yield=True,
                                    extract_independent=True)) == {yield1}
    assert set(gauss_ext.get_params(floating=None, is_yield=True,
                                    extract_independent=True)) == {yield1}
    assert set(gauss_ext.get_params(floating=False, is_yield=True,
                                    extract_independent=True)) == set()

    # Gauss ext2
    assert set(gauss2_ext.get_params(floating=True, is_yield=False, extract_independent=False)) == {mu2, sigma2}
    assert set(gauss2_ext.get_params(floating=None, is_yield=False, extract_independent=False)) == {mu2, sigma2}
    assert set(gauss2_ext.get_params(floating=False, is_yield=False, extract_independent=False)) == set()

    with pytest.raises(ValueError):
        set(gauss2_ext.get_params(floating=True, is_yield=None, extract_independent=False))
    assert set(gauss2_ext.get_params(floating=None, is_yield=None,
                                     extract_independent=False)) == {mu2, sigma2, yield2_comp}
    with pytest.raises(ValueError):
        set(gauss2_ext.get_params(floating=False, is_yield=None, extract_independent=False))

    with pytest.raises(ValueError):
        set(gauss2_ext.get_params(floating=False, is_yield=True, extract_independent=False))
    assert set(gauss2_ext.get_params(floating=None, is_yield=True, extract_independent=False)) == {yield2_comp}
    with pytest.raises(ValueError):
        set(gauss2_ext.get_params(floating=True, is_yield=True, extract_independent=False))

    # with extract deps
    assert set(gauss2_ext.get_params(floating=True, is_yield=False, extract_independent=True)) == {mu2, sigma2}
    assert set(gauss2_ext.get_params(floating=None, is_yield=False,
                                     extract_independent=True)) == {mu2, sigma2}
    yield2.floating = False
    assert set(gauss2_ext.get_params(floating=False, is_yield=False,
                                     extract_independent=True)) == set()
    yield2.floating = True
    assert set(gauss2_ext.get_params(floating=False, is_yield=False,
                                     extract_independent=True)) == set()

    assert set(gauss2_ext.get_params(floating=True, is_yield=None,
                                     extract_independent=True)) == {mu2, sigma2, yield2}
    assert set(gauss2_ext.get_params(floating=None, is_yield=None,
                                     extract_independent=True)) == {mu2, sigma2, yield2}
    assert set(gauss2_ext.get_params(floating=False, is_yield=None,
                                     extract_independent=True)) == set()

    assert set(gauss2_ext.get_params(floating=True, is_yield=True,
                                     extract_independent=True)) == {yield2}
    assert set(gauss2_ext.get_params(floating=None, is_yield=True,
                                     extract_independent=True)) == {yield2}
    assert set(gauss2_ext.get_params(floating=False, is_yield=True,
                                     extract_independent=True)) == set()

    # sum extended
    with pytest.raises(ValueError):
        assert set(sum1_ext.get_params(floating=True, is_yield=False, extract_independent=False))
    frac0 = sum1_ext.fracs[0]
    frac1 = sum1_ext.fracs[1]
    assert set(sum1_ext.get_params(floating=None, is_yield=False, extract_independent=False)) == {mu_nofloat,
                                                                                                  sigma_comp, mu2,
                                                                                                  sigma2, frac0, frac1}
    with pytest.raises(ValueError):
        assert set(sum1_ext.get_params(floating=False, is_yield=False, extract_independent=False))

    with pytest.raises(ValueError):
        set(sum1_ext.get_params(floating=True, is_yield=None, extract_independent=False))
    assert set(sum1_ext.get_params(floating=None, is_yield=None,
                                   extract_independent=False)) == {mu_nofloat, sigma_comp,
                                                                   mu2, sigma2, frac0, frac1,
                                                                   sum1_ext.get_yield()}
    with pytest.raises(ValueError):
        set(sum1_ext.get_params(floating=False, is_yield=None, extract_independent=False))

    with pytest.raises(ValueError):
        set(sum1_ext.get_params(floating=False, is_yield=True, extract_independent=False))
    assert set(sum1_ext.get_params(floating=None, is_yield=True, extract_independent=False)) == {sum1_ext.get_yield()}
    with pytest.raises(ValueError):
        set(sum1_ext.get_params(floating=True, is_yield=True, extract_independent=False))

    # with extract deps
    assert set(sum1_ext.get_params(floating=True, is_yield=False,
                                   extract_independent=True)) == {mu2,
                                                                  sigma2,
                                                                  yield1, yield2  # fracs depend on them
                                                                  }
    assert set(sum1_ext.get_params(floating=None, is_yield=False,
                                   extract_independent=True)) == {mu_nofloat, mu2, sigma2, yield1, yield2}
    assert set(sum1_ext.get_params(floating=False, is_yield=False,
                                   extract_independent=True)) == {mu_nofloat}

    assert set(sum1_ext.get_params(floating=True, is_yield=None,
                                   extract_independent=True)) == {mu2, sigma2, yield1, yield2, yield1, yield2}
    assert set(sum1_ext.get_params(floating=None, is_yield=None,
                                   extract_independent=True)) == {mu_nofloat, mu2, sigma2, yield1, yield2}
    assert set(sum1_ext.get_params(floating=False, is_yield=None,
                                   extract_independent=True)) == {mu_nofloat}
    yield1.floating = False
    assert set(sum1_ext.get_params(floating=False, is_yield=None,
                                   extract_independent=True)) == {mu_nofloat, yield1}
    yield1.floating = True

    assert set(sum1_ext.get_params(floating=True, is_yield=True,
                                   extract_independent=True)) == {yield1, yield2}
    assert set(sum1_ext.get_params(floating=None, is_yield=True,
                                   extract_independent=True)) == {yield1, yield2}
    assert set(sum1_ext.get_params(floating=False, is_yield=True,
                                   extract_independent=True)) == set()



"""Baseclass for most objects appearing in zfit."""
#  Copyright (c) 2021 zfit
import itertools
import warnings
from collections import OrderedDict
from typing import Iterable, Optional, Set

import tensorflow as tf
from ordered_set import OrderedSet

from ..util import ztyping
from ..util.cache import GraphCachable
from ..util.checks import NotSpecified
from ..util.container import convert_to_container
from .dependents import BaseDependentsMixin
from .interfaces import (ZfitIndependentParameter, ZfitNumericParametrized,
                         ZfitObject, ZfitParameter, ZfitParametrized)


class BaseObject(ZfitObject):

    def __init__(self, name, **kwargs):
        assert not kwargs, f"kwargs not empty, the following arguments are not captured: {kwargs}"
        super().__init__()

        self._name = name  # TODO: uniquify name?

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._repr = {}  # TODO: make repr more sophisticated

    @property
    def name(self) -> str:
        """The name of the object."""
        return self._name

    def copy(self, deep: bool = False, name: str = None, **overwrite_params) -> "ZfitObject":

        new_object = self._copy(deep=deep, name=name, overwrite_params=overwrite_params)
        return new_object

    def _copy(self, deep, name, overwrite_params):  # TODO(Mayou36) L: representation?
        raise NotImplementedError("This copy should not be used.")

    def __eq__(self, other: object) -> bool:
        if not isinstance(self, type(other)):
            return False
        try:
            for key, own_element in self._repr.items():
                if not own_element == other._repr.get(key):  # TODO: make repr better
                    return False
        except AttributeError:
            return self is other
        return True  # no break occurred

    def __hash__(self):
        return object.__hash__(self)


class BaseParametrized(ZfitParametrized):

    def __init__(self, params, **kwargs) -> None:
        super().__init__(**kwargs)
        from zfit.core.parameter import convert_to_parameter

        params = params or OrderedDict()
        params = OrderedDict(sorted((n, convert_to_parameter(p)) for n, p in params.items()))

        # parameters = OrderedDict(sorted(parameters))  # to always have a consistent order
        self._params = params
        self._repr['params'] = self.params

    def get_params(self,
                   floating: Optional[bool] = True,
                   is_yield: Optional[bool] = None,
                   extract_independent: Optional[bool] = True,
                   only_floating=NotSpecified) -> Set["ZfitParameter"]:
        """Recursively collect parameters that this object depends on according to the filter criteria.

        Which parameters should be included can be steered using the arguments as a filter.
         - **None**: do not filter on this. E.g. `floating=None` will return parameters that are floating as well as
            parameters that are fixed.
         - **True**: only return parameters that fulfil this criterion
         - **False**: only return parameters that do not fulfil this criterion. E.g. `floating=False` will return
            only parameters that are not floating.

        Args:
            floating: if a parameter is floating, e.g. if :py:meth:`~ZfitParameter.floating` returns `True`
            is_yield: if a parameter is a yield of the _current_ model. This won't be applied recursively, but may include
               yields if they do also represent a parameter parametrizing the shape. So if the yield of the current
               model depends on other yields (or also non-yields), this will be included. If, however, just submodels
               depend on a yield (as their yield) and it is not correlated to the output of our model, they won't be
               included.
            extract_independent: If the parameter is an independent parameter, i.e. if it is a `ZfitIndependentParameter`.
        """
        if only_floating is not NotSpecified:
            floating = only_floating
            warnings.warn("`only_floating` is deprecated and will be removed in the future, use `floating` instead.",
                          DeprecationWarning, stacklevel=2)
        return self._get_params(floating=floating, is_yield=is_yield, extract_independent=extract_independent)

    def _get_params(self, floating: Optional[bool] = True, is_yield: Optional[bool] = None,
                    extract_independent: Optional[bool] = True) -> Set["ZfitParameter"]:

        if is_yield is True:  # we want exclusively yields, we don't have them by default
            params = OrderedSet()
        else:
            params = self.params.values()
            params = extract_filter_params(params, floating=floating, extract_independent=extract_independent)
        return params

    @property
    def params(self) -> ztyping.ParameterType:
        return self._params


class BaseNumeric(GraphCachable, BaseDependentsMixin, BaseParametrized, ZfitNumericParametrized, BaseObject):

    def __init__(self, **kwargs):
        if 'dtype' in kwargs:  # TODO(Mayou36): proper dtype handling?
            self._dtype = kwargs.pop('dtype')
        super().__init__(**kwargs)
        self.add_cache_deps(self.params.values())

    @property
    def dtype(self) -> tf.DType:
        """The dtype of the object."""
        return self._dtype

    @staticmethod
    def _filter_floating_params(params):
        params = [param for param in params if isinstance(param, ZfitIndependentParameter) and param.floating]
        return params


def extract_filter_params(params: Iterable[ZfitParametrized],
                          floating: Optional[bool] = True,
                          extract_independent: Optional[bool] = True) -> Set[ZfitParameter]:
    params = convert_to_container(params, container=OrderedSet)

    if extract_independent:
        params = OrderedSet(itertools.chain.from_iterable(param.get_params(floating=floating,
                                                                           extract_independent=True,
                                                                           is_yield=None)
                                                          for param in params))

    if floating is not None:
        if not extract_independent and not all(param.independent for param in params):
            raise ValueError("Since `extract_dependent` is not set to True, there are maybe dependent parameters for "
                             "which `floating` is an ill-defined attribute.")
        params = OrderedSet(p for p in params if p.floating == floating)
    return params



#  Copyright (c) 2021 zfit
import warnings

from .core.loss import BaseLoss, ExtendedUnbinnedNLL, SimpleLoss, UnbinnedNLL

__all__ = ['ExtendedUnbinnedNLL', "UnbinnedNLL", "BaseLoss", "SimpleLoss", 'experimental_enable_loss_penalty']

from .util.warnings import warn_experimental_feature


@warn_experimental_feature
def experimental_enable_loss_penalty(enable=True):
    """EXPERIMENTAL! Enable a loss penalty if the loss is NaN, which can push back the minimizer.

    Won't work with every minimizer

    Args:
        enable: If True, enable this feature.
    """
    warnings.warn("This has been removed and is now activated by default. Remove this function call."
                  "Many thanks for the feedbacks received.")



"""Used to make pytest functions available globally."""

#  Copyright (c) 2021 zfit
import sys

import pytest

try:
    import pytest_randomly
except ImportError:
    pass
else:
    import zfit

    pytest_randomly.random_seeder = [zfit.settings.set_seed]
init_modules = sys.modules.keys()


@pytest.fixture(autouse=True)
def setup_teardown():
    import zfit
    old_chunksize = zfit.run.chunking.max_n_points
    old_active = zfit.run.chunking.active
    old_graph_mode = zfit.run.get_graph_mode()
    old_autograd_mode = zfit.run.get_autograd_mode()

    for m in sys.modules.keys():
        if m not in init_modules:
            del (sys.modules[m])

    yield
    from zfit.core.parameter import ZfitParameterMixin
    ZfitParameterMixin._existing_params.clear()

    from zfit.util.cache import clear_graph_cache
    clear_graph_cache()
    import zfit
    zfit.run.chunking.active = old_active
    zfit.run.chunking.max_n_points = old_chunksize
    zfit.run.set_graph_mode(old_graph_mode)
    zfit.run.set_autograd_mode(old_autograd_mode)
    for m in sys.modules.keys():
        if m not in init_modules:
            del (sys.modules[m])
    import gc
    gc.collect()


def pytest_addoption(parser):
    parser.addoption("--longtests", action="store", default=False)



#  Copyright (c) 2021 zfit

import tensorflow as tf

from zfit.settings import ztypes


def _auto_upcast(tensor: tf.Tensor):
    if isinstance(tensor, tf.Tensor):
        new_dtype = ztypes[tensor.dtype]
        if new_dtype != tensor.dtype:
            tensor = tf.cast(x=tensor, dtype=new_dtype)
    return tensor



#  Copyright (c) 2021 zfit

import numpy as np
import pytest

import zfit
import zfit.models.dist_tfp
import zfit.models.kde


@pytest.mark.skip()  # copy not yet implemented
def test_copy_kde():
    size = 500
    data = np.random.normal(size=size, loc=2, scale=3)

    limits = (-15, 5)
    obs = zfit.Space("obs1", limits=limits)
    kde_adaptive = zfit.models.kde.GaussianKDE1DimV1(data=data, bandwidth='adaptiveV1',
                                                     obs=obs,
                                                     truncate=False)
    kde_adaptive.copy()


def test_simple_kde():
    expected_integral = 5 / 6
    h = zfit.Parameter("h", 0.9)

    size = 5000
    data = np.random.normal(size=size, loc=2, scale=3)

    limits = (-15, 5)
    obs = zfit.Space("obs1", limits=limits)
    data_truncated = obs.filter(data)
    kde = zfit.models.kde.GaussianKDE1DimV1(data=data, bandwidth=h, obs=obs,
                                            truncate=False)
    kde_adaptive = zfit.models.kde.GaussianKDE1DimV1(data=data, bandwidth='adaptive',
                                                     obs=obs,
                                                     truncate=False)
    kde_silverman = zfit.models.kde.GaussianKDE1DimV1(data=data, bandwidth='silverman',
                                                      obs=obs,
                                                      truncate=False)
    kde_adaptive_trunc = zfit.models.kde.GaussianKDE1DimV1(data=data_truncated, bandwidth='adaptiveV1',
                                                           obs=obs,
                                                           truncate=True)

    integral = kde.integrate(limits=limits, norm_range=False)
    integral_trunc = kde_adaptive_trunc.integrate(limits=limits, norm_range=False)
    integral_adaptive = kde_adaptive.integrate(limits=limits, norm_range=False)
    integral_silverman = kde_silverman.integrate(limits=limits, norm_range=False)

    rel_tol = 0.04
    assert zfit.run(integral_trunc) == pytest.approx(1., rel=rel_tol)
    assert zfit.run(integral) == pytest.approx(expected_integral, rel=rel_tol)
    assert zfit.run(integral_adaptive) == pytest.approx(expected_integral, rel=rel_tol)
    assert zfit.run(integral_silverman) == pytest.approx(expected_integral, rel=rel_tol)

    sample = kde_adaptive.sample(1000)
    sample2 = kde_adaptive_trunc.sample(1500)
    prob = kde_adaptive.pdf(sample2)
    kde_adaptive_trunc.pdf(sample)
    assert prob.shape.rank == 1
    assert sample.nevents == 1000
    assert sample2.nevents == 1500



"""This example illustrates how to minimize an arbitrary Python function using the zfit minimizer.

This may has some overhead in the beginning and won't be instantly fast compared to other libraries if run once.

Copyright (c) 2021 zfit
"""

#  Copyright (c) 2021 zfit

import numpy as np

import zfit

# set everything to numpy mode
zfit.run.set_autograd_mode(False)
zfit.run.set_graph_mode(False)

# create our favorite minimizer
minimizer = zfit.minimize.IpyoptV1()


# minimizer = zfit.minimize.Minuit()
# minimizer = zfit.minimize.ScipyTrustConstrV1()
# minimizer = zfit.minimize.NLoptLBFGSV1()


def func(x):
    x = np.array(x)  # make sure it's an array
    return np.sum((x - 0.1) ** 2 + x[1] ** 4)


# we can also use a more complicated function instead
# from scipy.optimize import rosen as func


# we need to set the errordef, the definition of "1 sigma"
func.errordef = 0.5

# initial parameters
params = [1, -3, 2, 1.4, 11]
# or for a more fine-grained control
# params = {
#     'value': [1, -3, 2, 1.4, 11],  # mandatory
#     'lower': [-2, -5, -5, -10, -15],  # lower bound, can be omitted
#     'upper': [2, 4, 5, 10, 15],  # upper bound, can be omitted
#     'step_size': [0.1] * 5,  # initial step size, can be omitted
# }

# minimize
result = minimizer.minimize(func, params)

# estimate errors
result.hesse()
result.errors()
print(result)



#  Copyright (c) 2021 zfit

import matplotlib.pyplot as plt
import numpy as np

import zfit

n_bins = 50

# create space
obs = zfit.Space("x", limits=(-10, 10))

# parameters
mu = zfit.Parameter("mu", 1., -4, 6)
sigma = zfit.Parameter("sigma", 1., 0.1, 10)
lambd = zfit.Parameter("lambda", -0.06, -1, -0.01)

# model building, pdf creation
gauss = zfit.pdf.Gauss(mu=mu, sigma=sigma, obs=obs)
exponential = zfit.pdf.Exponential(lambd, obs=obs)

n_bkg = zfit.Parameter('n_bkg', 20000)
n_sig = zfit.Parameter('n_sig', 1000)
gauss_extended = gauss.create_extended(n_sig)
exp_extended = exponential.create_extended(n_bkg)
model = zfit.pdf.SumPDF([gauss_extended, exp_extended])

# data
# n_sample = 10000

data = model.create_sampler(n=21200)
data.resample()

# set the values to a start value for the fit
mu.set_value(0.5)
sigma.set_value(1.2)
lambd.set_value(-0.05)

# create NLL
nll = zfit.loss.ExtendedUnbinnedNLL(model=model, data=data)

# create a minimizer
minimizer = zfit.minimize.Minuit(verbosity=7)
result = minimizer.minimize(nll)
print(result.params)
# do the error calculations, here with hesse, than with minos
param_hesse = result.hesse()
param_errors, _ = result.errors()  # this returns a new FitResult if a new minimum was found
print(result.valid)  # check if the result is still valid



#  Copyright (c) 2021 zfit

import abc
from abc import ABCMeta, abstractmethod
from typing import Callable, Dict, List, Optional, Set, Tuple, Union

import numpy as np
import tensorflow as tf

from ..util import ztyping
from ..util.deprecation import deprecated


class ZfitObject(abc.ABC):

    # TODO: make abstractmethod?
    def __eq__(self, other: object) -> bool:
        raise NotImplementedError


class ZfitDimensional(ZfitObject):

    @property
    @abstractmethod
    def obs(self) -> ztyping.ObsTypeReturn:
        """Return the observables, string identifier for the coordinate system."""
        raise NotImplementedError

    @property
    @abstractmethod
    def axes(self) -> ztyping.AxesTypeReturn:
        """Return the axes, integer based identifier(indices) for the coordinate system."""
        raise NotImplementedError

    @property
    @abstractmethod
    def n_obs(self) -> int:
        """Return the number of observables, the dimensionality.

        Corresponds to the last dimension.
        """
        raise NotImplementedError


class ZfitOrderableDimensional(ZfitDimensional, metaclass=ABCMeta):

    @abstractmethod
    def with_obs(self,
                 obs: Optional[ztyping.ObsTypeInput],
                 allow_superset: bool = True,
                 allow_subset: bool = True) -> "ZfitOrderableDimensional":
        """Create a new instance that has `obs`; sorted by or set or dropped.

        The behavior is as follows:

         * obs are already set:
           * input obs are None: the observables will be dropped. If no axes are set, an error
             will be raised, as no coordinates will be assigned to this instance anymore.
           * input obs are not None: the instance will be sorted by the incoming obs. If axes or other
             objects have an associated order (e.g. data, limits,...), they will be reordered as well.
             If a strict subset is given (and allow_subset is True), only a subset will be returned.
             This can be used to take a subspace of limits, data etc.
             If a strict superset is given (and allow_superset is True), the obs will be sorted accordingly as
             if the obs not contained in the instances obs were not in the input obs.
         * obs are not set:
           * if the input obs are None, the same object is returned.
           * if the input obs are not None, they will be set as-is and now correspond to the already
             existing axes in the object.

        Args:
            obs: Observables to sort/associate this instance with
            allow_superset: if False and a strict superset of the own observables is given, an error
            is raised.
            allow_subset:if False and a strict subset of the own observables is given, an error
            is raised.

        Returns:
            A copy of the object with the new ordering/observables

        Raises:
            CoordinatesUnderdefinedError: if obs is None and the instance does not have axes
            ObsIncompatibleError: if `obs` is a superset and allow_superset is False or a subset and
                allow_allow_subset is False
        """
        raise NotImplementedError

    @abstractmethod
    def with_axes(self,
                  axes: Optional[ztyping.AxesTypeInput],
                  allow_superset: bool = True,
                  allow_subset: bool = True) -> "ZfitOrderableDimensional":
        """Create a new instance that has `axes`; sorted by or set or dropped.

        The behavior is as follows:

         * axes are already set:
           * input axes are None: the axes will be dropped. If no observables are set, an error
             will be raised, as no coordinates will be assigned to this instance anymore.
           * input axes are not None: the instance will be sorted by the incoming axes. If obs or other
             objects have an associated order (e.g. data, limits,...), they will be reordered as well.
             If a strict subset is given (and allow_subset is True), only a subset will be returned. This can
             be used to retrieve a subspace of limits, data etc.
             If a strict superset is given (and allow_superset is True), the axes will be sorted accordingly as
             if the axes not contained in the instances axes were not present in the input axes.
         * axes are not set:
           * if the input axes are None, the same object is returned.
           * if the input axes are not None, they will be set as-is and now correspond to the already
             existing obs in the object.

        Args:
            axes: Axes to sort/associate this instance with
            allow_superset: if False and a strict superset of the own axeservables is given, an error
            is raised.
            allow_subset:if False and a strict subset of the own axeservables is given, an error
            is raised.

        Returns:
            A copy of the object with the new ordering/axes

        Raises:
            CoordinatesUnderdefinedError: if obs is None and the instance does not have axes
            AxesIncompatibleError: if `axes` is a superset and allow_superset is False or a subset and
                allow_allow_subset is False
        """

        raise NotImplementedError

    @abstractmethod
    def with_autofill_axes(self, overwrite: bool = False) -> "ZfitOrderableDimensional":
        """Overwrite the axes of the current object with axes corresponding to range(len(n_obs)).

        This effectively fills with (0, 1, 2,...) and can be used mostly when an object enters a PDF or
        similar. `overwrite` allows to remove the axis first in case there are already some set.

        .. code-block::

            object.obs -> ('x', 'z', 'y')
            object.axes -> None

            object.with_autofill_axes()

            object.obs -> ('x', 'z', 'y')
            object.axes -> (0, 1, 2)


        Args:
            overwrite: If axes are already set, replace the axes with the autofilled ones.
                If axes is already set and `overwrite` is False, raise an error.

        Returns:
            The object with the new axes

        Raises:
            AxesIncompatibleError: if the axes are already set and `overwrite` is False.
        """
        raise NotImplementedError

    @abstractmethod
    def reorder_x(self, x: Union[tf.Tensor, np.ndarray],
                  *,
                  x_obs: ztyping.ObsTypeInput = None,
                  x_axes: ztyping.AxesTypeInput = None,
                  func_obs: ztyping.ObsTypeInput = None,
                  func_axes: ztyping.AxesTypeInput = None
                  ) -> ztyping.XTypeReturnNoData:
        """Reorder x in the last dimension either according to its own obs or assuming a function ordered with func_obs.

        There are two obs or axes around: the one associated with this Coordinate object and the one associated with x.
        If x_obs or x_axes is given, then this is assumed to be the obs resp. the axes of x and x will be reordered
        according to `self.obs` resp. `self.axes`.

        If func_obs resp. func_axes is given, then x is assumed to have `self.obs` resp. `self.axes` and will be
        reordered to align with a function ordered with `func_obs` resp. `func_axes`.

        Switching `func_obs` for `x_obs` resp. `func_axes` for `x_axes` inverts the reordering of x.

        Args:
            x: Tensor to be reordered, last dimension should be n_obs resp. n_axes
            x_obs: Observables associated with x. If both, x_obs and x_axes are given, this has precedency over the
                latter.
            x_axes: Axes associated with x.
            func_obs: Observables associated with a function that x will be given to. Reorders x accordingly and assumes
                self.obs to be the obs of x. If both, `func_obs` and `func_axes` are given, this has precedency over the
                latter.
            func_axes: Axe associated with a function that x will be given to. Reorders x accordingly and assumes
                self.axes to be the axes of x.

        Returns:
            The reordered array-like object
        """
        raise NotImplementedError

    @abstractmethod
    def get_reorder_indices(self,
                            obs: ztyping.ObsTypeInput = None,
                            axes: ztyping.AxesTypeInput = None
                            ) -> Tuple[int]:
        """Indices that would order the instances obs as `obs` respectively the instances axes as `axes`.

        Args:
            obs: Observables that the instances obs should be ordered to. Does not reorder, but just
                return the indices that could be used to reorder.
            axes: Axes that the instances obs should be ordered to. Does not reorder, but just
                return the indices that could be used to reorder.

        Returns:
            New indices that would reorder the instances obs to be obs respectively axes.

        Raises:
            CoordinatesUnderdefinedError: If neither `obs` nor `axes` is given
        """
        raise NotImplementedError


class ZfitData(ZfitDimensional):

    @abstractmethod
    def value(self, obs: List[str] = None) -> ztyping.XType:
        raise NotImplementedError

    @abstractmethod
    def sort_by_obs(self, obs, allow_superset: bool = True):
        raise NotImplementedError

    @abstractmethod
    def sort_by_axes(self, axes, allow_superset: bool = True):
        raise NotImplementedError

    @property
    @abstractmethod
    def weights(self):
        raise NotImplementedError


class ZfitLimit(abc.ABC, metaclass=ABCMeta):

    @property
    @abstractmethod
    def rect_limits(self) -> ztyping.RectLimitsReturnType:
        """Return the rectangular limits as `np.ndarray``tf.Tensor` if they are set and not false.

            The rectangular limits can be used for sampling. They do not in general represent the limits
            of the object as a functional limit can be set and to check if something is inside the limits,
            the method :py:meth:`~Limit.inside` should be used.

            In order to test if the limits are False or None, it is recommended to use the appropriate methods
            `limits_are_false` and `limits_are_set`.

        Returns:
            The lower and upper limits.
        Raises:
            LimitsNotSpecifiedError: If there are not limits set or they are False.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def rect_limits_np(self) -> ztyping.RectLimitsNPReturnType:
        """Return the rectangular limits as `np.ndarray`. Raise error if not possible.

        Rectangular limits are returned as numpy arrays which can be useful when doing checks that do not
        need to be involved in the computation later on as they allow direct interaction with Python as
        compared to `tf.Tensor` inside a graph function.

        In order to test if the limits are False or None, it is recommended to use the appropriate methods
        `limits_are_false` and `limits_are_set`.

        Returns:
            A tuple of two `np.ndarray` with shape (1, n_obs) typically. The last
                dimension is always `n_obs`, the first can be vectorized. This allows unstacking
                with `z.unstack_x()` as can be done with data.

        Raises:
            CannotConvertToNumpyError: In case the conversion fails.
            LimitsNotSpecifiedError: If the limits are not set or are false
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def rect_lower(self) -> ztyping.RectLowerReturnType:
        """The lower, rectangular limits, equivalent to `rect_limits[0]` with shape (..., n_obs)

        Returns:
            The lower, rectangular limits as `np.ndarray` or `tf.Tensor`
        Raises:
            LimitsNotSpecifiedError: If the limits are not set or are false
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def rect_upper(self) -> ztyping.RectUpperReturnType:
        """The upper, rectangular limits, equivalent to `rect_limits[1]` with shape (..., n_obs)

        Returns:
            The upper, rectangular limits as `np.ndarray` or `tf.Tensor`
        Raises:
            LimitsNotSpecifiedError: If the limits are not set or are false
        """
        raise NotImplementedError

    @abstractmethod
    def rect_area(self) -> Union[float, np.ndarray, tf.Tensor]:
        """Calculate the total rectangular area of all the limits and axes.

        Useful, for example, for MC integration.
        """
        raise NotImplementedError

    @abstractmethod
    def inside(self, x: ztyping.XTypeInput, guarantee_limits: bool = False) -> ztyping.XTypeReturn:
        """Test if `x` is inside the limits.

        This function should be used to test if values are inside the limits. If the given x is already inside
        the rectangular limits, e.g. because it was sampled from within them

        Args:
            x: Values to be checked whether they are inside of the limits. The shape is expected to have the last
                dimension equal to n_obs.
            guarantee_limits: Guarantee that the values are already inside the rectangular limits.

        Returns:
            Return a boolean tensor-like object with the same shape as the input `x` except of the
                last dimension removed.
        """
        raise NotImplementedError

    @abstractmethod
    def filter(self, x: ztyping.XTypeInput,
               guarantee_limits: bool = False,
               axis: Optional[int] = None
               ) -> ztyping.XTypeReturnNoData:
        """Filter `x` by removing the elements along `axis` that are not inside the limits.

        This is similar to `tf.boolean_mask`.

        Args:
            x: Values to be checked whether they are inside of the limits. If not, the corresonding element (in the
                specified `axis`) is removed. The shape is expected to have the last dimension equal to n_obs.
            guarantee_limits: Guarantee that the values are already inside the rectangular limits.
            axis: The axis to remove the elements from. Defaults to 0.

        Returns:
            Return an object with the same shape as `x` except that along `axis` elements have been
                removed.
        """

    @property
    @abstractmethod
    def has_rect_limits(self) -> bool:
        """If there are limits and whether they are rectangular."""
        raise NotImplementedError

    @property
    def rect_limits_are_tensors(self) -> bool:
        """Return True if the rectangular limits are tensors.

        If a limit with tensors is evaluated inside a graph context, comparison operations will fail.

        Returns:
            If the rectangular limits are tensors.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def limits_are_set(self) -> bool:
        """If the limits have been set to a limit or are False.

        Returns:
            Whether the limits have been set or not.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def limits_are_false(self) -> bool:
        """Returns if the limits have been set to False, so the object on purpose does not contain limits."""
        raise NotImplementedError

    @property
    @abstractmethod
    def has_limits(self) -> bool:
        """Whether there are limits set and they are not false."""
        raise NotImplementedError

    # TODO: remove from API?
    def get_subspace(self, *_, **__):
        from zfit.util.exception import InvalidLimitSubspaceError
        raise InvalidLimitSubspaceError("ZfitLimits does not suppoert subspaces")

    @property
    @abstractmethod
    def n_obs(self) -> int:
        """Dimensionality, the number of observables, of the limits. Equals to the last axis in rectangular limits.

        Returns:
            Dimensionality of the limits.
        """
        raise NotImplementedError

    @property
    def n_events(self) -> Union[int, None]:
        """Shape of the first dimension, usually reflects the number of events.

        Returns:
            Return the number of events, the dimension of the first shape. If this is > 1 or None,
                it's vectorized.
        """
        raise NotImplementedError

    @abstractmethod
    def equal(self, other: object, allow_graph: bool) -> Union[bool, tf.Tensor]:
        """Compare the limits on equality. For ANY objects, this also returns true.

        If called inside a graph context *and* the limits are tensors, this will return a symbolic `tf.Tensor`.

        Returns:
            Result of the comparison
        Raises:
             IllegalInGraphModeError: it the comparison happens with tensors in a graph context.
        """
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """Compares two Limits for equality without graph mode allowed.

        Raises:
             IllegalInGraphModeError: it the comparison happens with tensors in a graph context.
        """
        raise NotImplementedError

    @abstractmethod
    def less_equal(self,
                   other: object,
                   allow_graph: bool = True
                   ) -> Union[bool, tf.Tensor]:
        """Set-like comparison for compatibility. If an object is less_equal to another, the limits are combatible.

        This can be used to determine whether a fitting range specification can handle another limit.

        If called inside a graph context *and* the limits are tensors, this will return a symbolic `tf.Tensor`.


        Args:
            other: Any other object to compare with
            allow_graph: If False and the function returns a symbolic tensor, raise IllegalInGraphModeError instead.

        Returns:
            Result of the comparison
        Raises:
             IllegalInGraphModeError: it the comparison happens with tensors in a graph context.
        """
        raise NotImplementedError

    @abstractmethod
    def __le__(self, other: object) -> bool:
        """Set-like comparison for compatibility. If an object is less_equal to another, the limits are combatible.

        This can be used to determine whether a fitting range specification can handle another limit.

        Returns:
            Result of the comparison
        Raises:
             IllegalInGraphModeError: it the comparison happens with tensors in a graph context.
        """
        raise NotImplementedError

    @abstractmethod
    def get_sublimits(self):
        """Splits itself into multiple sublimits with smaller n_obs.

        If this is not possible, if the limits are not rectangular, just returns itself.

        Returns:
            The sublimits if it was able to split.
        """
        raise NotImplementedError


class ZfitSpace(ZfitLimit, ZfitOrderableDimensional, ZfitObject, metaclass=ABCMeta):

    @property
    @abstractmethod
    def n_limits(self) -> int:
        """Return the number of limits."""
        raise NotImplementedError

    # TODO: legacy?
    @property
    @abstractmethod
    def limits(self) -> Tuple[ztyping.LowerTypeReturn, ztyping.UpperTypeReturn]:
        """Return the tuple(lower, upper)."""
        raise NotImplementedError

    # TODO: legacy?
    @property
    @abstractmethod
    def lower(self) -> ztyping.LowerTypeReturn:
        """Return the lower limits."""
        raise NotImplementedError

    # TODO: legacy?
    @property
    @abstractmethod
    def upper(self) -> ztyping.UpperTypeReturn:
        """Return the upper limits."""
        raise NotImplementedError

    # TODO: legacy?
    @abstractmethod
    def area(self) -> float:
        """Return the total area of all the limits and axes.

        Useful, for example, for MC integration.
        """
        raise NotImplementedError

    @abstractmethod
    def with_limits(self,
                    limits: ztyping.LimitsTypeInput = None,
                    rect_limits: Optional[ztyping.RectLimitsInputType] = None,
                    name: Optional[str] = None
                    ) -> "ZfitSpace":
        """Return a copy of the space with the new `limits` (and the new `name`).

        Args:
            limits: Limits to use. Can be rectangular, a function (requires to also specify `rect_limits`
                or an instance of ZfitLimit.
            rect_limits: Rectangular limits that will be assigned with the instance
            name: Human readable name

        Returns:
            Copy of the current object with the new limits.
        """
        raise NotImplementedError

    @abstractmethod
    def get_subspace(self, obs, axes, name):
        """Create a :py:class:`~zfit.Space` consisting of only a subset of the `obs`/`axes` (only one allowed).

        Args:
            obs:
            axes:
            name:

        Returns:
        """
        raise NotImplementedError

    @abstractmethod
    def with_coords(self,
                    coords: ZfitOrderableDimensional,
                    allow_superset: bool = True,
                    allow_subset: bool = True) -> object:
        """Create a new :py:class:`~zfit.Space` with reordered observables and/or axes.

        The behavior is that _at least one coordinate (obs or axes) has to be set in both instances
        (the space itself or in `coords`). If both match, observables is taken as the defining coordinate.
        The space is sorted according to the defining coordinate and the other coordinate is sorted as well.
        If either the space did not have the "weaker coordinate" (e.g. both have observables, but only coords
        has axes), then the resulting Space will have both.
        If both have both coordinates, obs and axes, and sorting for obs results in non-matchin axes results
        in axes being dropped.

        Args:
            coords: An instance of :py:class:`Coordinates`
            allow_superset: If false and a strict superset is given, an error is raised
            allow_subset: If false and a strict subset is given, an error is raised

        Returns:
            :py:class:`~zfit.Space`:
        Raises:
            CoordinatesUnderdefinedError: if neither both obs or axes are specified.
            CoordinatesIncompatibleError: if `coords` is a superset and allow_superset is False or a subset and
                allow_allow_subset is False
        """
        raise NotImplementedError

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError

    @abstractmethod
    def __len__(self):
        raise NotImplementedError


class ZfitDependenciesMixin:
    @abstractmethod
    def get_cache_deps(self, only_floating: bool = True) -> ztyping.DependentsType:
        raise NotImplementedError

    @deprecated(date=None, instructions="Use `get_params` instead if you want to retrieve the "
                                        "independent parameters or `get_cache_deps` in case you need "
                                        "the numerical cache dependents (advanced).")
    def get_dependencies(self, only_floating: bool = True) -> ztyping.DependentsType:
        # raise BreakingAPIChangeError
        return self.get_cache_deps(only_floating=only_floating)


class ZfitParametrized(ZfitDependenciesMixin, ZfitObject):
    @abstractmethod
    def get_params(self,
                   floating: Optional[bool] = True,
                   is_yield: Optional[bool] = None,
                   extract_independent: Optional[bool] = True
                   ) -> Set["ZfitParameter"]:
        """Recursively collect parameters that this object depends on according to the filter criteria.

        Which parameters should be included can be steered using the arguments as a filter.
         - **None**: do not filter on this. E.g. `floating=None` will return parameters that are floating as well as
            parameters that are fixed.
         - **True**: only return parameters that fulfil this criterion
         - **False**: only return parameters that do not fulfil this criterion. E.g. `floating=False` will return
            only parameters that are not floating.

        Args:
            floating: if a parameter is floating, e.g. if :py:meth:`~ZfitParameter.floating` returns `True`
            is_yield: if a parameter is a yield of the _current_ model. This won't be applied recursively, but may include
               yields if they do also represent a parameter parametrizing the shape. So if the yield of the current
               model depends on other yields (or also non-yields), this will be included. If, however, just submodels
               depend on a yield (as their yield) and it is not correlated to the output of our model, they won't be
               included.
            extract_independent: If the parameter is an independent parameter, i.e. if it is a `ZfitIndependentParameter`.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def params(self) -> ztyping.ParameterType:
        raise NotImplementedError


class ZfitNumericParametrized(ZfitParametrized):

    @property
    @abstractmethod
    def dtype(self) -> tf.DType:
        """The `DType` of `Tensor`s handled by this `model`."""
        raise NotImplementedError


class ZfitParameter(ZfitNumericParametrized):

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    # TODO: maybe add to numerics?
    @property
    @abstractmethod
    def shape(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def floating(self) -> bool:
        raise NotImplementedError

    @floating.setter
    @abstractmethod
    def floating(self, value: bool):
        raise NotImplementedError

    @abstractmethod
    def value(self) -> tf.Tensor:
        raise NotImplementedError

    @abstractmethod
    def read_value(self) -> tf.Tensor:
        raise NotImplementedError

    @property
    @abstractmethod
    def independent(self) -> bool:
        raise NotImplementedError


class ZfitIndependentParameter(ZfitParameter, metaclass=ABCMeta):
    @abstractmethod
    def randomize(self, minval, maxval, sampler):
        """Update the parameter with a randomised value between minval and maxval and return it.

        Args:
            minval: The lower bound of the sampler. If not given, `lower_limit` is used.
            maxval: The upper bound of the sampler. If not given, `upper_limit` is used.
            sampler: A sampler with the same interface as `tf.random.uniform`

        Returns:
            The sampled value
        """
        raise NotImplementedError

    @abstractmethod
    def set_value(self, value):
        """Set the :py:class:`~zfit.Parameter` to `value` (temporarily if used in a context manager).

        This operation won't, compared to the assign, return the read value but an object that *can* act as a context
        manager.

        Args:
            value: The value the parameter will take on.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def has_limits(self) -> bool:
        """If the parameter has limits set or not."""
        raise NotImplementedError

    @property
    @abstractmethod
    def at_limit(self) -> tf.Tensor:
        """If the value is at the limit (or over it).

        Returns:
            Boolean `tf.Tensor` that tells whether the value is at the limits.
        """
        raise NotImplementedError

    @property
    def step_size(self) -> tf.Tensor:
        """Step size of the parameter, the estimated order of magnitude of the uncertainty.

        This can be crucial to tune for the minimization. A too large `step_size` can produce NaNs, a too small won't
        converge.

        If the step size is not set, the `DEFAULT_STEP_SIZE` is used.

        Returns:
            The step size
        """
        raise NotImplementedError


class ZfitLoss(ZfitObject, metaclass=ABCMeta):

    @abstractmethod
    def gradient(self, params: ztyping.ParamTypeInput = None) -> List[tf.Tensor]:
        raise NotImplementedError

    @abstractmethod
    def value(self) -> ztyping.NumericalTypeReturn:
        raise NotImplementedError

    @property
    @abstractmethod
    def model(self) -> List["ZfitModel"]:
        raise NotImplementedError

    @property
    @abstractmethod
    def data(self) -> List["ZfitData"]:
        raise NotImplementedError

    @property
    @abstractmethod
    def fit_range(self) -> List["ZfitSpace"]:
        raise NotImplementedError

    @abstractmethod
    def add_constraints(self, constraints: List[tf.Tensor]):
        raise NotImplementedError

    @property
    @abstractmethod
    def errordef(self) -> float:
        raise NotImplementedError

    def hessian(self, params):
        pass

    @abstractmethod
    def value_gradient(self, params):
        pass

    @abstractmethod
    def value_gradient_hessian(self, params, hessian=None):
        pass

    @abstractmethod
    def create_new(self, **kwargs):
        pass


class ZfitModel(ZfitNumericParametrized, ZfitDimensional):

    @abstractmethod
    def update_integration_options(self, *args, **kwargs):  # TODO: handling integration properly
        raise NotImplementedError

    @abstractmethod
    def integrate(self, limits: ztyping.LimitsType, norm_range: ztyping.LimitsType = None,
                  name: str = "integrate") -> ztyping.XType:
        """Integrate the function over `limits` (normalized over `norm_range` if not False).

        Args:
            limits: the limits to integrate over
            norm_range: the limits to normalize over or False to integrate the
                unnormalized probability
            name:

        Returns:
            The integral value
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def register_analytic_integral(cls, func: Callable, limits: ztyping.LimitsType = None,
                                   priority: int = 50, *,
                                   supports_norm_range: bool = False,
                                   supports_multiple_limits: bool = False):
        """Register an analytic integral with the class.

        Args:
            func:
            limits: |limits_arg_descr|
            priority:
            supports_multiple_limits:
            supports_norm_range:

        Returns:
        """
        raise NotImplementedError

    @abstractmethod
    def partial_integrate(self, x: ztyping.XType, limits: ztyping.LimitsType,
                          norm_range: ztyping.LimitsType = None) -> ztyping.XType:
        """Partially integrate the function over the `limits` and evaluate it at `x`.

        Dimension of `limits` and `x` have to add up to the full dimension and be therefore equal
        to the dimensions of `norm_range` (if not False)

        Args:
            x: The value at which the partially integrated function will be evaluated
            limits: the limits to integrate over. Can contain only some axes
            norm_range: the limits to normalize over. Has to have all axes

        Returns:
            The value of the partially integrated function evaluated at `x`.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def register_inverse_analytic_integral(cls, func: Callable):
        """Register an inverse analytical integral, the inverse (unnormalized) cdf.

        Args:
            func:
        """
        raise NotImplementedError

    @abstractmethod
    def sample(self, n: int, limits: ztyping.LimitsType = None) -> ztyping.XType:
        """Sample `n` points within `limits` from the model.

        Args:
            n: The number of samples to be generated
            limits: In which region to sample in
            name:

        Returns:
            Tensor(n_obs, n_samples)
        """
        raise NotImplementedError


class ZfitFunc(ZfitModel):
    @abstractmethod
    def func(self, x: ztyping.XType, name: str = "value") -> ztyping.XType:
        raise NotImplementedError

    @abstractmethod
    def as_pdf(self):
        raise NotImplementedError


class ZfitPDF(ZfitModel):

    @abstractmethod
    def pdf(self, x: ztyping.XType, norm_range: ztyping.LimitsType = None) -> ztyping.XType:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_extended(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def set_norm_range(self):
        raise NotImplementedError

    @abstractmethod
    def create_extended(self, yield_: ztyping.ParamTypeInput) -> "ZfitPDF":
        raise NotImplementedError

    @abstractmethod
    def get_yield(self) -> Union[ZfitParameter, None]:
        raise NotImplementedError

    @abstractmethod
    def normalization(self, limits: ztyping.LimitsType) -> ztyping.NumericalTypeReturn:
        raise NotImplementedError

    @abstractmethod
    def as_func(self, norm_range: ztyping.LimitsType = False):
        raise NotImplementedError


class ZfitFunctorMixin:

    @property
    @abstractmethod
    def models(self) -> Dict[Union[float, int, str], ZfitModel]:
        raise NotImplementedError

    @abstractmethod
    def get_models(self) -> List[ZfitModel]:
        raise NotImplementedError


class ZfitConstraint(abc.ABC):
    @abstractmethod
    def value(self):
        raise NotImplementedError



#  Copyright (c) 2021 zfit

import numpy as np
import pytest

import zfit
from zfit.core.sample import extended_sampling, extract_extended_pdfs

obs1 = zfit.Space('obs1', limits=(-3, 4))


@pytest.mark.flaky(reruns=3)  # poissonian sampling
def test_extract_extended_pdfs():
    gauss1 = zfit.pdf.Gauss(obs=obs1, mu=1.3, sigma=5.4)
    gauss2 = zfit.pdf.Gauss(obs=obs1, mu=1.3, sigma=5.4)
    gauss3 = zfit.pdf.Gauss(obs=obs1, mu=1.3, sigma=5.4)
    gauss4 = zfit.pdf.Gauss(obs=obs1, mu=1.3, sigma=5.4)
    gauss5 = zfit.pdf.Gauss(obs=obs1, mu=1.3, sigma=5.4)
    gauss6 = zfit.pdf.Gauss(obs=obs1, mu=1.3, sigma=5.4)

    yield1 = zfit.Parameter('yield123' + str(np.random.random()), 200.)

    # sum1 = 0.3 * gauss1 + gauss2
    gauss3_ext = gauss3.create_extended(45)
    gauss4_ext = gauss4.create_extended(100)
    sum2_ext_daughters = gauss3_ext + gauss4_ext
    sum3 = zfit.pdf.SumPDF((gauss5, gauss6), 0.4)
    sum3_ext = sum3.create_extended(yield1)

    sum_all = zfit.pdf.SumPDF(pdfs=[sum2_ext_daughters, sum3_ext])
    sum_all.set_norm_range((-5, 5))

    extracted_pdfs = extract_extended_pdfs(pdfs=sum_all)
    assert frozenset(extracted_pdfs) == {gauss3_ext, gauss4_ext, sum3_ext}

    limits = zfit.Space(obs=obs1, limits=(-4, 5))
    limits = limits.with_autofill_axes()
    extended_sample = extended_sampling(pdfs=sum_all, limits=limits)
    extended_sample_np = extended_sample.numpy()
    assert np.shape(extended_sample_np)[0] == pytest.approx(expected=(45 + 100 + 200), rel=0.1)
    samples_from_pdf = sum_all.sample(n='extended', limits=limits)
    samples_from_pdf_np = samples_from_pdf.numpy()
    assert np.shape(samples_from_pdf_np)[0] == pytest.approx(expected=(45 + 100 + 200), rel=0.1)


def test_set_yield():
    gauss6 = zfit.pdf.Gauss(obs=obs1, mu=1.3, sigma=5.4)

    yield1 = zfit.Parameter('yield123' + str(np.random.random()), 200.)
    assert not gauss6.is_extended
    gauss6.set_yield(yield1)
    assert gauss6.is_extended



#  Copyright (c) 2021 zfit
import pytest

import zfit
from zfit.util.exception import NormRangeUnderdefinedError

limits1 = (-4, 3)
limits2 = (-2, 5)
limits3 = (-1, 7)
obs1 = 'obs1'
obs2 = 'obs2'

space1 = zfit.Space(obs=obs1, limits=limits1)
space2 = zfit.Space(obs=obs1, limits=limits2)
space3 = zfit.Space(obs=obs1, limits=limits3)
space4 = zfit.Space(obs=obs2, limits=limits2)

space5 = space1.combine(space4)


def test_norm_range():
    gauss1 = zfit.pdf.Gauss(1., 4., obs=space1)
    gauss2 = zfit.pdf.Gauss(1., 4., obs=space1)
    gauss3 = zfit.pdf.Gauss(1., 4., obs=space2)

    sum1 = zfit.pdf.SumPDF(pdfs=[gauss1, gauss2], fracs=0.4, obs=space1)
    assert sum1.obs == (obs1,)
    assert sum1.norm_range == space1

    with pytest.raises(NormRangeUnderdefinedError):
        _ = zfit.pdf.SumPDF(pdfs=[gauss1, gauss3], fracs=0.34)
    sum2 = zfit.pdf.SumPDF(pdfs=[gauss1, gauss3], fracs=0.34, obs=space3)

    sum2.set_norm_range(space2)
    with sum2.set_norm_range(space3):
        assert sum2.norm_range == space3
    assert sum2.norm_range == space2


def test_combine_range():
    gauss1 = zfit.pdf.Gauss(1., 4., obs=space1)
    gauss4 = zfit.pdf.Gauss(1., 4., obs=space4)
    gauss5 = zfit.pdf.Gauss(1., 4., obs=space4)

    product = zfit.pdf.ProductPDF(pdfs=[gauss1, gauss4])
    assert product.obs == (obs1, obs2)
    assert product.norm_range == space5

    product = zfit.pdf.ProductPDF(pdfs=[gauss1, gauss4, gauss5])
    assert product.obs == (obs1, obs2)
    assert product.norm_range == space5



#  Copyright (c) 2021 zfit

import abc
import inspect
import warnings
from typing import (Callable, Iterable, List, Mapping, Optional, Set, Tuple,
                    Union)

import tensorflow as tf
from ordered_set import OrderedSet

import zfit.z.numpy as znp

from .. import settings, z
from ..util import ztyping
from ..util.checks import NONE
from ..util.container import convert_to_container, is_container
from ..util.deprecation import deprecated, deprecated_args
from ..util.exception import (BehaviorUnderDiscussion, BreakingAPIChangeError,
                              IntentionAmbiguousError, NotExtendedPDFError)
from ..util.warnings import warn_advanced_feature
from ..z.math import (autodiff_gradient, autodiff_value_gradients,
                      automatic_value_gradients_hessian, numerical_gradient,
                      numerical_value_gradient,
                      numerical_value_gradients_hessian)
from .baseobject import BaseNumeric, extract_filter_params
from .constraint import BaseConstraint
from .dependents import _extract_dependencies
from .interfaces import ZfitData, ZfitLoss, ZfitPDF, ZfitSpace
from .parameter import convert_to_parameters, set_values


# @z.function
def _unbinned_nll_tf(model: ztyping.PDFInputType, data: ztyping.DataInputType, fit_range: ZfitSpace, log_offset=None):
    """Return unbinned negative log likelihood graph for a PDF.

    Args:
        model: PDFs with a `.pdf` method. Has to be as many models as data
        data:
        fit_range:

    Returns:
        The unbinned nll

    Raises:
        ValueError: if both `probs` and `log_probs` are specified.
    """

    if is_container(model):
        nlls = [_unbinned_nll_tf(model=p, data=d, fit_range=r, log_offset=log_offset)
                for p, d, r in zip(model, data, fit_range)]
        # nlls_total = [nll.total for nll in nlls]
        # nlls_correction = [nll.correction for nll in nlls]
        # nlls_total_summed = znp.sum(input_tensor=nlls_total, axis=0)
        nlls_summed = znp.sum(nlls, axis=0)

        # nlls_correction_summed = znp.sum(input_tensor=nlls_correction, axis=0)
        # nll_finished = (nlls_total_summed, nlls_correction_summed)
        nll_finished = nlls_summed
    else:
        if fit_range is not None:
            with data.set_data_range(fit_range):
                probs = model.pdf(data, norm_range=fit_range)
        else:
            probs = model.pdf(data)
        log_probs = znp.log(probs)
        nll = _nll_calc_unbinned_tf(log_probs=log_probs,
                                    weights=data.weights if data.weights is not None else None,
                                    log_offset=log_offset)
        nll_finished = nll
    return nll_finished


@z.function(wraps='tensor')
def _nll_calc_unbinned_tf(log_probs, weights=None, log_offset=None):
    if weights is not None:
        log_probs *= weights  # because it's prob ** weights
    if log_offset is not None:
        log_probs -= log_offset
    nll = -znp.sum(log_probs, axis=0)
    # nll = -tfp.math.reduce_kahan_sum(input_tensor=log_probs, axis=0)
    return nll


def _constraint_check_convert(constraints):
    checked_constraints = []
    for constr in constraints:
        if isinstance(constr, BaseConstraint):
            checked_constraints.append(constr)
        else:
            raise BreakingAPIChangeError("Constraints have to be of type `Constraint`, a simple"
                                         " constraint from a function can be constructed with"
                                         " `SimpleConstraint`.")
    return checked_constraints


class BaseLoss(ZfitLoss, BaseNumeric):

    def __init__(self, model: ztyping.ModelsInputType, data: ztyping.DataInputType,
                 fit_range: ztyping.LimitsTypeInput = None,
                 constraints: ztyping.ConstraintsTypeInput = None,
                 options: Optional[Mapping] = None):
        # first doc line left blank on purpose, subclass adds class docstring (Sphinx autodoc adds the two)
        """A "simultaneous fit" can be performed by giving one or more `model`, `data`, `fit_range` to the loss. The
        length of each has to match the length of the others.

        Args:
            model: The model or models to evaluate the data on
            data: Data to use
            fit_range: The fitting range. It's the norm_range for the models (if
            they
                have a norm_range) and the data_range for the data.
            constraints: A Tensor representing a loss constraint. Using
                `zfit.constraint.*` allows for easy use of predefined constraints.
            options: Different options for the loss calculation.
        """
        super().__init__(name=type(self).__name__, params={})
        if fit_range is not None and all(fr is not None for fr in fit_range):
            warnings.warn("The fit_range argument is depreceated and will maybe removed in future releases. "
                          "It is preferred to define the range in the space"
                          " when creating the data and the model.", stacklevel=2)

        model, data, fit_range = self._input_check(pdf=model, data=data, fit_range=fit_range)
        self._model = model
        self._data = data
        self._fit_range = fit_range

        options = self._check_init_options(options, data)

        self._options = options
        self._subtractions = {}
        if constraints is None:
            constraints = []
        self._constraints = _constraint_check_convert(convert_to_container(constraints, list))

        self._precompile()

    def _check_init_options(self, options, data):
        try:
            nevents = sum(d.nevents for d in data)
        except RuntimeError:  # can happen if not yet sampled. What to do? Approx_nevents?
            nevents = 150_000  # sensible default
        options = {} if options is None else options

        if options.get('numhess') is None:
            options['numhess'] = True

        if options.get('numgrad') is None:
            options['numgrad'] = settings.options['numerical_grad']

        if options.get('kahansum') is None:
            options['kahansum'] = nevents > 500_000  # start using kahan if we have more than 500k events

        if options.get('subtr_const') is None:  # TODO: balance better?

            # if nevents < 200_000:
            #     subtr_const = True
            # elif nevents < 1_000_000:
            #     subtr_const = 'kahan'
            # else:
            #     subtr_const = 'elewise'
            subtr_const = True
            options['subtr_const'] = subtr_const

        return options

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._name = "UnnamedSubBaseLoss"

    def _get_params(self,
                    floating: Optional[bool] = True,
                    is_yield: Optional[bool] = None,
                    extract_independent: Optional[bool] = True) -> Set["ZfitParameter"]:
        params = OrderedSet()
        params = params.union(*(model.get_params(floating=floating, is_yield=is_yield,
                                                 extract_independent=extract_independent)
                                for model in self.model))

        params = params.union(*(constraint.get_params(floating=floating, is_yield=False,
                                                      extract_independent=extract_independent)
                                for constraint in self.constraints))
        return params

    def _input_check(self, pdf, data, fit_range):

        if isinstance(pdf, tuple):
            raise TypeError("`pdf` has to be a pdf or a list of pdfs, not a tuple.")
        if isinstance(data, tuple):
            raise TypeError("`data` has to be a data or a list of data, not a tuple.")

        # pdf, data = (convert_to_container(obj, non_containers=[tuple]) for obj in (pdf, data))
        pdf, data = self._check_convert_model_data(pdf, data, fit_range)
        # TODO: data, range consistency?
        if fit_range is None:
            fit_range = []
            non_consistent = {'data': [], 'model': [], 'range': []}
            for p, d in zip(pdf, data):
                if p.norm_range != d.data_range:
                    non_consistent['data'].append(d)
                    non_consistent['model'].append(p)
                    non_consistent['range'].append((p.space, d.data_range))
                fit_range.append(None)
            if non_consistent['range']:  # TODO: test
                warn_advanced_feature(f"PDFs {non_consistent['model']} as "
                                      f"well as `data` {non_consistent['data']}"
                                      f" have different ranges {non_consistent['range']} they"
                                      f" are defined in. The data range will cut the data while the"
                                      f" norm range defines the normalization.",
                                      identifier='inconsistent_fitrange')
        else:
            fit_range = convert_to_container(fit_range, non_containers=[tuple])

        if not len(pdf) == len(data) == len(fit_range):
            raise ValueError("pdf, data and fit_range don't have the same number of components:"
                             "\npdf: {}"
                             "\ndata: {}"
                             "\nfit_range: {}".format(pdf, data, fit_range))

        # sanitize fit_range
        fit_range = [p.convert_sort_space(limits=range_) if range_ is not None else None for p, range_ in
                     zip(pdf, fit_range)]
        # TODO: sanitize pdf, data?
        self.add_cache_deps(cache_deps=pdf)
        self.add_cache_deps(cache_deps=data)
        return pdf, data, fit_range

    def _precompile(self):
        do_subtr = self._options.get('subtr_const', False)
        if do_subtr:
            if do_subtr is not True:
                self._options['subtr_const_value'] = do_subtr
            log_offset = self._options.get('subtr_const_value')
            if log_offset is None:
                from zfit import run
                run.assert_executing_eagerly()  # first time subtr
                nevents_tot = znp.sum([d._approx_nevents for d in self.data])
                log_offset_sum = (self._call_value(data=self.data,
                                                   model=self.model,
                                                   fit_range=self.fit_range,
                                                   constraints=self.constraints,
                                                   # presumably were not at the minimum,
                                                   # so the loss will decrease
                                                   log_offset=z.convert_to_tensor(0.)) - 1000.)
                log_offset = tf.stop_gradient(- znp.divide(log_offset_sum, nevents_tot))
                self._options['subtr_const_value'] = log_offset

    def _check_convert_model_data(self, model, data, fit_range):
        model, data = tuple(convert_to_container(obj) for obj in (model, data))

        model_checked = []
        data_checked = []
        for mod, dat in zip(model, data):
            if not isinstance(dat, ZfitData):
                if fit_range is not None:
                    raise TypeError("Fit range should not be used if data is not ZfitData.")

                if not isinstance(dat, (tf.Tensor, tf.Variable)):
                    try:
                        dat = z.convert_to_tensor(value=dat)
                    except TypeError:
                        raise TypeError(
                            f"Wrong type of dat ({type(dat)}). Has to be a `ZfitData` or convertible to a tf.Tensor")
                # check dimension
                from zfit import Data
                dat = Data.from_tensor(obs=mod.space, tensor=dat)
            model_checked.append(mod)
            data_checked.append(dat)
        return model_checked, data_checked

    def _input_check_params(self, params):
        if params is None:
            params = list(self.get_params())
        else:
            params = convert_to_container(params)
        return params

    def add_constraints(self, constraints):
        constraints = convert_to_container(constraints)
        return self._add_constraints(constraints)

    def _add_constraints(self, constraints):
        constraints = _constraint_check_convert(convert_to_container(constraints, container=list))
        self._constraints.extend(constraints)
        return constraints

    @property
    def name(self):
        return self._name

    @property
    def model(self):
        return self._model

    @property
    def data(self):
        return self._data

    @property
    def fit_range(self):
        return self._fit_range

    @property
    def constraints(self):
        return self._constraints

    def _get_dependencies(self):  # TODO: fix, add constraints
        pdf_dependents = _extract_dependencies(self.model)
        pdf_dependents |= _extract_dependencies(self.constraints)
        return pdf_dependents

    @abc.abstractmethod
    def _loss_func(self, model, data, fit_range, constraints, log_offset):
        raise NotImplementedError

    @property
    def errordef(self) -> Union[float, int]:
        return self._errordef

    def __call__(self, _x: ztyping.DataInputType = None) -> tf.Tensor:
        """Calculate the loss value with the given input for the free parameters.

        Args:
            *positional*: Array-like argument to set the parameters. The order of the values correspond to
                the position of the parameters in :py:meth:`~BaseLoss.get_params()` (called without any arguments).
                For more detailed control, it is always possible to wrap :py:meth:`~BaseLoss.value()` and set the
                desired parameters manually.

        Returns:
            Calculated loss value as a scalar.
        """
        if _x is None:
            raise BehaviorUnderDiscussion("Currently, calling a loss requires to give the arguments explicitly."
                                          " If you think this behavior should be changed, please open an issue"
                                          " https://github.com/zfit/zfit/issues/new/choose")
        if isinstance(_x, dict):
            raise TypeError("Dicts are not supported when calling a loss, only array-like values.")
        if _x is None:
            return self.value()
        else:
            params = self.get_params()
            with set_values(params, _x):
                return self.value()

    def value(self):
        log_offset = self._options.get('subtr_const_value')

        value = self._call_value(self.model, self.data, self.fit_range, self.constraints, log_offset)
        return value

    def _call_value(self, model, data, fit_range, constraints, log_offset):
        value = self._value(model=model,
                            data=data,
                            fit_range=fit_range,
                            constraints=constraints,
                            log_offset=log_offset)
        # if self._subtractions.get('kahan') is None:
        #     self._subtractions['kahan'] = value
        # value_subtracted = (value[0] - self._subtractions['kahan'][0]) - (
        #         value[1] - self._subtractions['kahan'][1])
        # return value_subtracted
        return value
        # value = value_substracted[0] - value_substracted[1]

    def _value(self, model, data, fit_range, constraints, log_offset):
        # try:
        return self._loss_func(model=model, data=data, fit_range=fit_range,
                               constraints=constraints, log_offset=log_offset)

    # except NotImplementedError as error:
    #     raise NotImplementedError(f"_loss_func not properly defined! error {error}") from error

    def __add__(self, other):
        if not isinstance(other, BaseLoss):
            raise TypeError("Has to be a subclass of `BaseLoss` or overwrite `__add__`.")
        if type(other) != type(self):
            raise ValueError("cannot safely add two different kind of loss.")
        model = self.model + other.model
        data = self.data + other.data
        fit_range = self.fit_range + other.fit_range
        constraints = self.constraints + other.constraints
        return type(self)(
            model=model, data=data, fit_range=fit_range, constraints=constraints
        )

    def gradient(self, params: ztyping.ParamTypeInput = None) -> List[tf.Tensor]:
        params = self._input_check_params(params)
        numgrad = self._options['numgrad']

        return self._gradient(params=params, numgrad=numgrad)

    @deprecated(None, "Use `gradient` instead.")
    def gradients(self, *args, **kwargs):
        return self.gradient(*args, **kwargs)

    # @z.function(wraps='loss')
    def _gradient(self, params, numgrad):
        if numgrad:
            return numerical_gradient(self.value, params=params)

        else:
            return autodiff_gradient(self.value, params=params)

    def value_gradient(self, params: ztyping.ParamTypeInput) -> Tuple[tf.Tensor, tf.Tensor]:
        params = self._input_check_params(params)
        numgrad = self._options['numgrad']
        return self._value_gradient(params=params, numgrad=numgrad)

    @deprecated(None, "Use `value_gradient` instead.")
    def value_gradients(self, *args, **kwargs):
        return self.value_gradient(*args, **kwargs)

    @z.function(wraps='loss')
    def _value_gradient(self, params, numgrad=False):
        if numgrad:
            value, gradient = numerical_value_gradient(self.value, params=params)
        else:
            value, gradient = autodiff_value_gradients(self.value, params=params)
        return value, gradient

    def hessian(self, params: ztyping.ParamTypeInput, hessian=None):
        return self.value_gradient_hessian(params=params, hessian=hessian)[2]

    def value_gradient_hessian(self, params: ztyping.ParamTypeInput, hessian=None, numgrad=None) -> Tuple[
        tf.Tensor, tf.Tensor, tf.Tensor]:
        params = self._input_check_params(params)
        numgrad = self._options['numhess'] if numgrad is None else numgrad
        vals = self._value_gradient_hessian(params=params, hessian=hessian, numerical=numgrad)

        vals = vals[0], z.convert_to_tensor(vals[1]), vals[2]
        return vals

    @deprecated(None, "Use `value_gradient_hessian` instead.")
    def value_gradients_hessian(self, *args, **kwargs):
        return self.value_gradient_hessian(*args, **kwargs)

    @z.function(wraps='loss')
    def _value_gradient_hessian(self, params, hessian, numerical=False):
        if numerical:
            return numerical_value_gradients_hessian(
                func=self.value,
                gradient=self.gradient,
                params=params, hessian=hessian)
        else:
            return automatic_value_gradients_hessian(
                self.value, params=params, hessian=hessian
            )

    def __repr__(self) -> str:
        class_name = repr(self.__class__)[:-2].split(".")[-1]
        return f'<{class_name} ' \
               f'model={one_two_many([model.name for model in self.model])} ' \
               f'data={one_two_many([data.name for data in self.data])} ' \
               f'constraints={one_two_many(self.constraints, many="True")} ' \
               f'>'

    def __str__(self) -> str:
        class_name = repr(self.__class__)[:-2].split(".")[-1]
        return f'<{class_name}' \
               f' model={one_two_many([model for model in self.model])}' \
               f' data={one_two_many([data for data in self.data])}' \
               f' constraints={one_two_many(self.constraints, many="True")}' \
               f'>'


def one_two_many(values, n=3, many='multiple'):
    values = convert_to_container(values)
    if len(values) > n:
        values = many
    return values


class UnbinnedNLL(BaseLoss):
    _name = "UnbinnedNLL"

    def __init__(self,
                 model: Union[ZfitPDF, Iterable[ZfitPDF]],
                 data: Union[ZfitData, Iterable[ZfitData]],
                 fit_range=None,
                 constraints=None,
                 options=None):
        """Unbinned Negative Log Likelihood.

        A simultaneous fit can be performed by giving one or more `model`, `data`, `fit_range` to the loss. The
        length of each has to match the length of the others.

        Args:
            model: If not given, the current one will be used.
                |@doc:loss.init.model| PDFs that return the normalized probability for
               *data* under the given parameters.
               If multiple model and data are given, they will be used
               in the same order to do a simultaneous fit. |@docend:loss.init.model|
            data: If not given, the current one will be used.
                |@doc:loss.init.data| Dataset that will be given to the *model*.
               If multiple model and data are given, they will be used
               in the same order to do a simultaneous fit. |@docend:loss.init.data|
            fit_range: The fitting range. It's the norm_range for the models (if
                they have a norm_range) and the data_range for the data.
            constraints: If not given, the current one will be used.
                |@doc:loss.init.constraints| Auxiliary measurements that add a term to the loss
               or terms that restrict the loss in an other way such as penalties. |@docend:loss.init.constraints|
            options: If not given, the current one will be used.
                |@doc:loss.init.options| Additional options (as a dict) for the loss.
               Current possibilities include:

               - 'subtr_const' (default True): subtract from each points
                 log probability density a constant that
                 is approximately equal to the average log probability
                 density in the very first evaluation before
                 the summation. This brings the initial loss value closer to 0 and increases,
                 especially for large datasets, the numerical stability.

                 The value will be stored ith 'subtr_const_value' and can also be given
                 directly.

                 The subtraction should not affect the minimum as the absolute
                 value of the NLL is meaningless. However,
                 with this switch on, one cannot directly compare
                 different likelihoods ablolute value as the constant
                 may differs! Use `create_new` in order to have a comparable likelihood
                 between different losses


               These settings may extend over time. In order to make sure that a loss is the
               same under the same data, make sure to use `create_new` instead of instantiating
               a new loss as the former will automatically overtake any relevant constants
               and behavior. |@docend:loss.init.options|
        """
        super().__init__(model=model, data=data, fit_range=fit_range, constraints=constraints, options=options)
        self._errordef = 0.5
        extended_pdfs = [pdf for pdf in self.model if pdf.is_extended]
        if extended_pdfs and type(self) == UnbinnedNLL:
            warn_advanced_feature("Extended PDFs are given to a normal UnbinnedNLL. This won't take the yield "
                                  "into account and simply treat the PDFs as non-extended PDFs. To create an "
                                  "extended NLL, use the `ExtendedUnbinnedNLL`.", identifier='extended_in_UnbinnedNLL')

    def _loss_func(self, model, data, fit_range, constraints, log_offset):

        return self._loss_func_watched(data=data,
                                       model=model,
                                       fit_range=fit_range,
                                       constraints=constraints,
                                       log_offset=log_offset)

    @property
    def is_extended(self):
        return False

    @z.function(wraps='loss')
    def _loss_func_watched(self, data, model, fit_range, constraints, log_offset):
        nll = _unbinned_nll_tf(model=model, data=data,
                               fit_range=fit_range,
                               log_offset=log_offset)
        if constraints:
            constraints = z.reduce_sum([c.value() for c in constraints])
            nll += constraints
        return nll

    def _get_params(self, floating: Optional[bool] = True, is_yield: Optional[bool] = None,
                    extract_independent: Optional[bool] = True) -> Set["ZfitParameter"]:
        if not self.is_extended:
            is_yield = False  # the loss does not depend on the yields
        return super()._get_params(floating, is_yield, extract_independent)

    def create_new(self,
                   model: Optional[Union[ZfitPDF, Iterable[ZfitPDF]]] = NONE,
                   data: Optional[Union[ZfitData, Iterable[ZfitData]]] = NONE,
                   fit_range=NONE,
                   constraints=NONE,
                   options=NONE):
        """Create a new loss from the current loss and replacing what is given as the arguments.

        This creates a "copy" of the current loss but replacing any argument that is explicitly given.
        Equivalent to creating a new instance but with some arguments taken.

        A loss has more than a model and data (and constraints), it can have internal optimizations
        and more that may do alter the behavior of a naive re-instantiation in unpredictable ways.

        Args:
            model: If not given, the current one will be used.
                |@doc:loss.init.model| PDFs that return the normalized probability for
               *data* under the given parameters.
               If multiple model and data are given, they will be used
               in the same order to do a simultaneous fit. |@docend:loss.init.model|
            data: If not given, the current one will be used.
                |@doc:loss.init.data| Dataset that will be given to the *model*.
               If multiple model and data are given, they will be used
               in the same order to do a simultaneous fit. |@docend:loss.init.data|
            fit_range:
            constraints: If not given, the current one will be used.
                |@doc:loss.init.constraints| Auxiliary measurements that add a term to the loss
               or terms that restrict the loss in an other way such as penalties. |@docend:loss.init.constraints|
            options: If not given, the current one will be used.
                |@doc:loss.init.options| Additional options (as a dict) for the loss.
               Current possibilities include:

               - 'subtr_const' (default True): subtract from each points
                 log probability density a constant that
                 is approximately equal to the average log probability
                 density in the very first evaluation before
                 the summation. This brings the initial loss value closer to 0 and increases,
                 especially for large datasets, the numerical stability.

                 The value will be stored ith 'subtr_const_value' and can also be given
                 directly.

                 The subtraction should not affect the minimum as the absolute
                 value of the NLL is meaningless. However,
                 with this switch on, one cannot directly compare
                 different likelihoods ablolute value as the constant
                 may differs! Use `create_new` in order to have a comparable likelihood
                 between different losses


               These settings may extend over time. In order to make sure that a loss is the
               same under the same data, make sure to use `create_new` instead of instantiating
               a new loss as the former will automatically overtake any relevant constants
               and behavior. |@docend:loss.init.options|

        Returns:
        """
        if model is NONE:
            model = self.model
        if data is NONE:
            data = self.data
        if fit_range is NONE:
            fit_range = self.fit_range
        if constraints is NONE:
            constraints = self.constraints
            if constraints is not None:
                constraints = constraints.copy()
        if options is NONE:
            options = self._options
            if isinstance(options, dict):
                options = options.copy()
        return type(self)(model=model, data=data, fit_range=fit_range, constraints=constraints, options=options)


class ExtendedUnbinnedNLL(UnbinnedNLL):
    """An Unbinned Negative Log Likelihood with an additional poisson term for the number of events in the dataset."""

    @z.function(wraps='loss')
    def _loss_func(self, model, data, fit_range, constraints, log_offset):
        nll = super()._loss_func(model=model, data=data, fit_range=fit_range, constraints=constraints,
                                 log_offset=log_offset)
        yields = []
        nevents_collected = []
        for mod, dat in zip(model, data):
            if not mod.is_extended:
                raise NotExtendedPDFError(f"The pdf {mod} is not extended but has to be (for an extended fit)")
            nevents = dat.n_events if dat.weights is None else z.reduce_sum(dat.weights)
            nevents = tf.cast(nevents, tf.float64)
            nevents_collected.append(nevents)
            yields.append(mod.get_yield())
        yields = znp.stack(yields, axis=0)
        nevents_collected = znp.stack(nevents_collected, axis=0)

        term_new = tf.nn.log_poisson_loss(nevents_collected, znp.log(yields))
        if log_offset is not None:
            term_new += log_offset
        nll += znp.sum(term_new, axis=0)
        return nll

    @property
    def is_extended(self):
        return True


class SimpleLoss(BaseLoss):
    _name = "SimpleLoss"

    @deprecated_args(None, "Use params instead.", ('deps', 'dependents'))
    def __init__(self,
                 func: Callable,
                 params: Iterable["zfit.Parameter"] = None,
                 errordef: Optional[float] = None,
                 # legacy
                 deps: Iterable["zfit.Parameter"] = NONE,
                 dependents: Iterable["zfit.Parameter"] = NONE,
                 ):
        """Loss from a (function returning a) Tensor.

        This allows for a very generic loss function as the functions only restriction is that is
        should depend on `zfit.Parameter`.

        Args:
            func: Callable that constructs the loss and returns a tensor without taking an argument.
            params: The dependents (independent `zfit.Parameter`) of the loss. Essentially the (free) parameters that
              the `func` depends on.
            errordef: Definition of which change in the loss corresponds to a change of 1 sigma.
                For example, 1 for Chi squared, 0.5 for negative log-likelihood.

        Usage:

        .. code:: python

            import zfit
            import zfit.z.numpy as znp

            import tensorflow as tf

            param1 = zfit.Parameter('param1', 5, 1, 10)
            # we can build a model here if we want, but in principle, it's not necessary

            x = znp.random.uniform(size=(100,))
            # Todo: change to znp.random.normal when introduced into tensorflow.experimental.numpy.
            y = x*tf.random.normal(mean=4, stddev=0.1, shape=x.shape, dtype=tf.double)


            def squared_loss():
                y_pred = x*param1  # this is very simple, but we can of course use any
                # zfit PDF or Func inside
                squared = (y_pred - y)**2
                mse = znp.mean(squared)
                return mse

            squared_loss.errordef = 1

            loss = zfit.loss.SimpleLoss(squared_loss, param1)

        which can then be used in conjunction with any zfit minimizer such as Minuit

        .. code:: python

            minimizer = zfit.minimize.Minuit()
            result = minimizer.minimize(loss)
        """
        super().__init__(model=[], data=[], options={'subtr_const': False})
        if dependents is not NONE and params is None:
            params = dependents
        elif deps is not NONE and params is None:  # depreceation
            params = deps
        elif params is None:  # legacy, remove in 0.7
            raise BreakingAPIChangeError("params need to be specified explicitly due to the upgrade to 0.4."
                                         "More information can be found in the upgrade guide on the website.")

        # @z.function(wraps='loss')
        # def wrapped_func():
        #     return func()
        if hasattr(func, 'errordef'):
            errordef = func.errordef

        if errordef is None:
            raise ValueError(f"{self} cannot minimize {func} as `errordef` is missing: "
                             f"it has to be set as an attribute. Typically 1 (chi2) or 0.5 (NLL).")

        sig = inspect.signature(func)
        self._call_with_args = len(sig.parameters) > 0

        self._simple_func = func
        self._errordef = errordef
        params = convert_to_parameters(params, prefer_constant=False)
        self._params = params
        self._simple_func_params = _extract_dependencies(params)

    def _get_dependencies(self):
        dependents = self._simple_func_params
        return dependents

    def _get_params(self, floating: Optional[bool] = True, is_yield: Optional[bool] = None,
                    extract_independent: Optional[bool] = True) -> Set["ZfitParameter"]:
        params = super()._get_params(floating, is_yield, extract_independent)
        own_params = extract_filter_params(self._params, floating=floating, extract_independent=extract_independent)
        params = params.union(own_params)
        return params

    @property
    def errordef(self):
        errordef = self._errordef
        if errordef is None:
            raise RuntimeError("For this SimpleLoss, no error calculation is possible.")
        else:
            return errordef

    # @z.function(wraps='loss')
    def _loss_func(self, model, data, fit_range, constraints=None, log_offset=None):

        if self._call_with_args:
            params = self._simple_func_params

            value = self._simple_func(params)
        else:
            value = self._simple_func()
        return z.convert_to_tensor(value)

    def __add__(self, other):
        raise IntentionAmbiguousError("Cannot add a SimpleLoss, 'addition' of losses can mean anything."
                                      "Add them manually")

    def create_new(self,
                   func: Callable = NONE,
                   params: Iterable["zfit.Parameter"] = NONE,
                   errordef: Optional[float] = NONE
                   ):
        if func is NONE:
            func = self._simple_func
        if params is NONE:
            params = self._params
        if errordef is NONE:
            errordef = self.errordef

        return type(self)(func=func, params=params, errordef=errordef)



#  Copyright (c) 2021 zfit
import logging
from functools import lru_cache, wraps
from typing import Callable, Dict, List, Optional, Tuple, Union

import numdifftools
import numpy as np
import scipy.stats
import tensorflow as tf
from scipy import optimize

import zfit.z.numpy as znp

from .. import settings, z
from ..core.interfaces import ZfitIndependentParameter
from ..core.parameter import assign_values
from ..param import set_values
from ..util.container import convert_to_container
from ..util.deprecation import deprecated_args


class NewMinimum(Exception):
    """Exception class for cases where a new minimum is found."""
    pass


class FailEvalLossNaN(Exception):
    pass


@deprecated_args(None, "Use cl for confidence level instead.", 'sigma')
def compute_errors(result: "zfit.result.FitResult",
                   params: List[ZfitIndependentParameter],
                   cl: Optional[float] = None,
                   rtol: Optional[float] = 0.001,
                   method: Optional[str] = None,
                   covariance_method: Optional[Union[str, Callable]] = None,
                   sigma: float = 1,
                   ) -> Tuple[Dict[ZfitIndependentParameter, Dict[str, float]],
                              Union["zfit.result.FitResult", None]]:
    """Compute asymmetric errors of parameters by profiling the loss function in the fit result.

    This method finds the value for a given parameter where the loss function is `cl` away: for example
    for a cl of 68.3%, this is one (multiplied by the errordef). The other parameters are also minimized and
    not fixed. This method is comparably computationally intensive and, if possible, `hesse` should be used.
    However, since `hesse` does not capture asymetric or non-parabolic shaped profiles well, this method is
    preferable.

    Args:
        result: fit result to be used to compute the uncertainties.
        params: The parameters to calculate the
            errors error. If None, use all parameters.
        cl: Confidence Level of the parameter to be determined. Defaults to 68.3%.
        rtol: relative tol between the computed and the exact roots
        method: type of solver, `method` argument of :py:func:`scipy.optimize.root`. Defaults to "hybr".
        covariance_method: The method to use to calculate the correlation matrix, will be forwarded directly
            to :py:meth:`FitResult.covariance`. Valid choices are
            by default {'minuit_hesse', 'hesse_np'} (or any other method defined in the result)
            or a Callable.
        sigma: Errors are calculated with respect to `sigma` std deviations.


    Returns:
        out:
            A `dict` containing as keys the parameter and as value a `dict` which
            contains two keys 'lower' and 'upper', holding the calculated errors.
            Example: result[par1]['upper'] -> the asymmetric upper error of 'par1'
        out: a fit result is returned when a new minimum is found during the loss scan
    """
    method = "hybr" if method is None else method
    # method = "krylov" if method is None else method  # TODO: integration tests, better for large n params?
    if cl is None:
        factor = 1.0
    else:
        factor = scipy.stats.chi2(1).ppf(cl)
    params = convert_to_container(params)
    new_result = None

    all_params = list(result.params.keys())
    loss = result.loss
    errordef = loss.errordef * factor
    fmin = result.fmin
    rtol *= errordef
    minimizer = result.minimizer
    from zfit import run
    old_values = run(result)

    covariance = result.covariance(method=covariance_method, as_dict=True)
    param_errors = {param: covariance[(param, param)] ** 0.5 for param in params}
    # param_scale = np.array(list(param_errors.values()))  # TODO: can be used for root finding initialization?

    ncalls = 0
    try:
        # start = time.time()
        to_return = {}
        for param in params:
            assign_values(all_params, result)

            logging.info(f"profiling the parameter {param}")
            param_error = param_errors[param]
            param_value = result.params[param]["value"]
            other_params = [p for p in all_params if p != param]

            initial_values = {"lower": [], "upper": []}
            direction = {"lower": -sigma, "upper": sigma}

            for ap in all_params:
                ap_value = result.params[ap]["value"]
                error_factor = covariance[(param, ap)] * (2 * errordef / param_error ** 2) ** 0.5
                for d in ["lower", "upper"]:
                    ap_value_init = ap_value + direction[d] * error_factor
                    initial_values[d].append(ap_value_init)

            # TODO: improvement, use jacobian?
            # @np_cache(maxsize=25)
            def func(values, args):
                nonlocal ncalls
                ncalls += 1
                swap_sign = args

                assign_values(all_params, values)
                try:
                    loss_value, gradient = loss.value_gradient(params=other_params)
                except tf.errors.InvalidArgumentError:
                    msg = (f"The evaluation of the errors of {param.name} failed due to too many NaNs"
                           " being produced in the loss and/or its gradient. This is most probably"
                           " caused by negative values returned from the PDF.")
                    raise FailEvalLossNaN(msg)

                zeroed_loss = loss_value.numpy() - fmin

                gradient = np.array(gradient)
                if swap_sign(param):  # mirror at x-axis to remove second zero
                    zeroed_loss = - zeroed_loss
                    gradient = - gradient
                    logging.info("Swapping sign in error calculation 'zfit_error'")

                elif zeroed_loss < - minimizer.tol:
                    assign_values(all_params, values)  # set values to the new minimum
                    raise NewMinimum("A new minimum is found.")

                downward_shift = errordef * sigma ** 2
                shifted_loss = zeroed_loss - downward_shift

                return np.concatenate([[shifted_loss], gradient])

            to_return[param] = {}
            swap_sign = {
                "lower": lambda p: p > param_value,
                "upper": lambda p: p < param_value,
            }
            for d in ["lower", "upper"]:
                roots = optimize.root(fun=func,
                                      args=(swap_sign[d],),
                                      x0=np.array(initial_values[d]),
                                      tol=rtol,
                                      options={
                                          'factor': 0.1,
                                          # 'diag': 1 / param_scale,  # scale factor for variables
                                          # 'diag': param_scale,
                                      },
                                      method=method)
                to_return[param][d] = roots.x[all_params.index(param)] - param_value
                # print(f"error {d}, time needed {time.time() - start2}")
        # print(f"errors found, time needed {time.time() - start}")
        assign_values(all_params, old_values)

    except NewMinimum as e:
        from .. import settings
        if settings.get_verbosity() >= 5:
            print(e)
        minimizer = result.minimizer
        loss = result.loss
        new_found_fmin = loss.value()
        new_result = minimizer.minimize(loss=loss)
        if new_result.fmin >= new_found_fmin:
            raise RuntimeError("A new minimum was discovered but the minimizer was not able to find this on himself. "
                               "This behavior is currently an exception but will most likely change in the future.")
        to_return, new_result_ = compute_errors(result=new_result, params=params, sigma=sigma, rtol=rtol,
                                                method=method)
        if new_result_ is not None:
            new_result = new_result_
    return to_return, new_result


def numerical_pdf_jacobian(func, params):
    jacobian_func = numdifftools.Jacobian(func)
    return jacobian_func([param.value() for param in params]).T


@z.function(wraps='autodiff')
def autodiff_pdf_jacobian(func, params):
    columns = []

    for p in params:
        vector = np.zeros(len(params))
        vector[params.index(p)] = 1.
        with tf.autodiff.ForwardAccumulator(params, list(vector)) as acc:
            values = func()
        columns.append(acc.jvp(values))

    jacobian = z.convert_to_tensor(columns)

    return jacobian


def covariance_with_weights(method, result, params):
    model = result.loss.model
    data = result.loss.data
    from zfit import run
    old_vals = run(params)

    Hinv_dict = method(result=result, params=params)  # inverse of the hessian matrix
    if not Hinv_dict:
        return {}
    Hinv = dict_to_matrix(params, Hinv_dict)

    def func():
        values = []
        for m, d in zip(model, data):
            v = m.log_pdf(d)
            if d.weights is not None:
                v *= d.weights
            values.append(v)
        return znp.concatenate(values, axis=0)

    if settings.options['numerical_grad']:
        def wrapped_func(values):
            assign_values(params, values)
            return func()

        jacobian = numerical_pdf_jacobian(func=wrapped_func, params=params)
    else:
        jacobian = autodiff_pdf_jacobian(func=func, params=params).numpy()

    C = np.matmul(jacobian, jacobian.T)
    covariance = np.matmul(Hinv, np.matmul(C, Hinv))
    assign_values(params, old_vals)
    return matrix_to_dict(params, covariance)


def dict_to_matrix(params, matrix_dict):
    nparams = len(params)
    matrix = np.empty((nparams, nparams))
    if not matrix_dict:
        return None

    for i in range(nparams):
        pi = params[i]
        for j in range(i, nparams):
            pj = params[j]
            matrix[i, j] = matrix_dict[(pi, pj)]
            if i != j:
                matrix[j, i] = matrix_dict[(pi, pj)]

    return matrix


def matrix_to_dict(params, matrix):
    nparams = len(params)
    matrix_dict = {}
    if matrix is None:
        return matrix_dict

    for i in range(nparams):
        pi = params[i]
        for j in range(i, nparams):
            pj = params[j]
            matrix_dict[(pi, pj)] = matrix[i, j]
            if i != j:
                matrix_dict[(pj, pi)] = matrix[i, j]

    return matrix_dict


def np_cache(*args, **kwargs):
    """LRU cache implementation for functions whose FIRST parameter is a numpy array.

    >>> array = np.array([[1, 2, 3], [4, 5, 6]])
    >>> @np_cache(maxsize=256)
    ... def multiply(array, factor):
    ...     print("Calculating...")
    ...     return factor*array
    >>> multiply(array, 2)
    Calculating...
    array([[ 2,  4,  6],
           [ 8, 10, 12]])
    >>> multiply(array, 2)
    array([[ 2,  4,  6],
           [ 8, 10, 12]])
    >>> multiply.cache_info()
    CacheInfo(hits=1, misses=1, maxsize=256, currsize=1)
    """

    def decorator(function):
        @wraps(function)
        def wrapper(np_array, *args, **kwargs):
            hashable_array = array_to_tuple(np_array)
            return cached_wrapper(hashable_array, *args, **kwargs)

        @lru_cache(*args, **kwargs)
        def cached_wrapper(hashable_array, *args, **kwargs):
            array = np.array(hashable_array)
            return function(array, *args, **kwargs)

        def array_to_tuple(np_array):
            """Iterates recursively."""
            try:
                return tuple(array_to_tuple(_) for _ in np_array)
            except TypeError:
                return np_array

        # copy lru_cache attributes over too
        wrapper.cache_info = cached_wrapper.cache_info
        wrapper.cache_clear = cached_wrapper.cache_clear

        return wrapper

    return decorator



#  Copyright (c) 2021 zfit
import numpy as np
import pytest
import tensorflow as tf
from ordered_set import OrderedSet

import zfit
from zfit import Parameter, z
from zfit.core.parameter import ComplexParameter, ComposedParameter
from zfit.util.exception import NameAlreadyTakenError


def test_complex_param():
    real_part = 1.3
    imag_part = 0.3
    complex_value = real_part + 1j * imag_part

    param1 = ComplexParameter("param1_compl", lambda: complex_value, params=None)
    some_value = 3. * param1 ** 2 - 1.2j
    true_value = 3. * complex_value ** 2 - 1.2j
    assert true_value == pytest.approx(some_value.numpy(), rel=1e-8)
    assert not param1.get_params()

    # Cartesian complex
    real_part_param = Parameter("real_part_param", real_part)
    imag_part_param = Parameter("imag_part_param", imag_part)
    param2 = ComplexParameter.from_cartesian("param2_compl", real_part_param, imag_part_param)
    part1, part2 = param2.get_params()
    part1_val, part2_val = [part1.value().numpy(), part2.value().numpy()]
    if part1_val == pytest.approx(real_part):
        assert part2_val == pytest.approx(imag_part)
    elif part2_val == pytest.approx(real_part):
        assert part1_val == pytest.approx(imag_part)
    else:
        assert False, "one of the if or elif should be the case"

    # Polar complex
    mod_val = 2
    arg_val = np.pi / 4.0
    mod_part_param = Parameter("mod_part_param", mod_val)
    arg_part_param = Parameter("arg_part_param", arg_val)
    param3 = ComplexParameter.from_polar("param3_compl", mod_part_param, arg_part_param)
    part1, part2 = param3.get_cache_deps()
    part1_val, part2_val = [part1.value().numpy(), part2.value().numpy()]
    if part1_val == pytest.approx(mod_val):
        assert part2_val == pytest.approx(arg_val)
    elif part1_val == pytest.approx(arg_val):
        assert part2_val == pytest.approx(mod_val)
    else:
        assert False, "one of the if or elif should be the case"

    param4_name = "param4"
    param4 = ComplexParameter.from_polar(param4_name, 4., 2., floating=True)
    deps_param4 = param4.get_cache_deps()
    assert len(deps_param4) == 2
    for dep in deps_param4:
        assert dep.floating

    # Test complex conjugate
    param5 = param2.conj

    # Test properties (1e-8 is too precise)
    assert zfit.run(param1.real) == pytest.approx(real_part, rel=1e-6)
    assert zfit.run(param1.imag) == pytest.approx(imag_part, rel=1e-6)
    assert zfit.run(param2.real) == pytest.approx(real_part, rel=1e-6)
    assert zfit.run(param2.imag) == pytest.approx(imag_part, rel=1e-6)
    assert zfit.run(param2.mod) == pytest.approx(np.abs(complex_value))
    assert zfit.run(param2.arg) == pytest.approx(np.angle(complex_value))
    assert zfit.run(param3.mod) == pytest.approx(mod_val, rel=1e-6)
    assert zfit.run(param3.arg) == pytest.approx(arg_val, rel=1e-6)
    assert zfit.run(param3.real) == pytest.approx(mod_val * np.cos(arg_val), rel=1e-6)
    assert zfit.run(param3.imag) == pytest.approx(mod_val * np.sin(arg_val), rel=1e-6)
    assert zfit.run(param5.real) == pytest.approx(real_part)
    assert zfit.run(param5.imag) == pytest.approx(-imag_part)


def test_repr():
    val = 1543
    val2 = 1543 ** 2
    param1 = Parameter('param1', val)
    param2 = zfit.ComposedParameter('comp1', lambda x: x ** 2, params=param1)
    repr_value = repr(param1)
    repr_value2 = repr(param2)
    assert str(val) in repr_value
    assert str(val) in repr_value2

    @z.function
    def tf_call():
        repr_value = repr(param1)
        repr_value2 = repr(param2)
        assert str(val) not in repr_value
        assert str(val2) not in repr_value2
        assert 'graph-node' in repr_value
        assert 'graph-node' in repr_value2

    if zfit.run.get_graph_mode():  # only test if running in graph mode
        tf_call()


def test_composed_param():
    param1 = Parameter('param1', 1.)
    param2 = Parameter('param2', 2.)
    param3 = Parameter('param3', 3., floating=False)
    param4 = Parameter('param4', 4.)  # noqa Needed to make sure it does not only take all params as deps

    def value_fn(p1, p2, p3):
        return z.math.log(3. * p1) * tf.square(p2) - p3

    param_a = ComposedParameter('param_as', value_fn=value_fn, params=(param1, param2, param3))
    param_a2 = ComposedParameter('param_as2', value_fn=value_fn, params={f'p{i}': p
                                                                         for i, p in
                                                                         enumerate((param1, param2, param3))})
    assert param_a2.params['p1'] == param2
    assert isinstance(param_a.get_cache_deps(only_floating=True), OrderedSet)
    assert param_a.get_cache_deps(only_floating=True) == {param1, param2}
    assert param_a.get_cache_deps(only_floating=False) == {param1, param2, param3}
    a_unchanged = value_fn(param1, param2, param3).numpy()
    assert a_unchanged == param_a.numpy()
    param2.assign(3.5)
    assert param2.numpy()
    a_changed = value_fn(param1, param2, param3).numpy()
    assert a_changed == param_a.numpy()
    assert a_changed != a_unchanged

    print(param_a)

    @z.function
    def print_param(p):
        print(p)

    print_param(param_a)

    # TODO(params): reactivate to check?
    # with pytest.raises(LogicalUndefinedOperationError):
    #     param_a.assign(value=5.)
    # with pytest.raises(LogicalUndefinedOperationError):
    #     param_a.assign(value=5.)


def test_shape_parameter():
    a = Parameter(name='a', value=1)
    assert a.shape.rank == 0


def test_shape_composed_parameter():
    a = Parameter(name='a', value=1)
    b = Parameter(name='b', value=2)

    def compose():
        return tf.square(a) - b

    c = ComposedParameter(name='c', value_fn=compose, params=[a, b])
    assert c.shape.rank == 0


# TODO: add test
def test_randomize():
    param1 = zfit.Parameter('param1', 1.0, 0, 2)
    for _ in range(100):
        param1.randomize()
        assert 0 < param1 < 2


def test_floating_behavior():
    param1 = zfit.Parameter('param1', 1.0)
    assert param1.floating


def test_param_limits():
    lower, upper = -4., 3.
    param1 = Parameter('param1', 1., lower=lower, upper=upper)
    param2 = Parameter('param2', 2.)

    assert param1.has_limits
    assert not param2.has_limits

    with pytest.raises(ValueError):
        param1.set_value(upper + 0.5)
    param1.assign(upper + 0.5)
    assert upper == param1.value().numpy()
    assert param1.at_limit
    with pytest.raises(ValueError):
        param1.set_value(lower - 1.1)
    param1.assign(lower - 1.1)
    assert lower == param1.value().numpy()
    assert param1.at_limit
    param1.set_value(upper - 0.1)
    assert not param1.at_limit

    param2.lower = lower
    param2.assign(lower - 1.1)
    assert lower == param2.value().numpy()


def test_overloaded_operators():
    param1 = zfit.Parameter('param1', 5)
    param2 = zfit.Parameter('param2', 3)
    param_a = ComposedParameter('param_a', lambda p1: p1 * 4, params=param1)
    param_b = ComposedParameter('param_b', lambda p2: p2, params=param2)
    param_c = param_a * param_b
    assert not isinstance(param_c, zfit.Parameter)
    param_d = ComposedParameter("param_d", lambda pa, pb: pa + pa * pb ** 2, params=[param_a, param_b])
    param_d_val = param_d.numpy()
    assert param_d_val == (param_a + param_a * param_b ** 2).numpy()


# @pytest.mark.skip  # TODO: reactivate, causes segfault
def test_equal_naming():
    param_unique_name = zfit.Parameter('fafdsfds', 5.)
    with pytest.raises(NameAlreadyTakenError):
        param_unique_name2 = zfit.Parameter('fafdsfds', 3.)


# @pytest.mark.skip  # TODO: segfaulting?
def test_set_value():
    value1 = 1.
    value2 = 2.
    value3 = 3.
    value4 = 4.
    param1 = zfit.Parameter(name="param1", value=value1)
    assert param1.numpy() == value1
    with param1.set_value(value2):
        assert param1.numpy() == value2
        param1.set_value(value3)
        assert param1.numpy() == value3
        with param1.set_value(value4):
            assert param1.numpy() == value4
        assert param1.numpy() == value3
    assert param1.numpy() == value1


def test_fixed_param():
    obs = zfit.Space("obs1", (-1, 2))
    sigma = zfit.param.ConstantParameter('const1', 5)
    gauss = zfit.pdf.Gauss(1., sigma, obs=obs)
    mu = gauss.params['mu']
    assert isinstance(mu, zfit.param.ConstantParameter)
    assert isinstance(sigma, zfit.param.ConstantParameter)
    assert not sigma.floating
    assert not sigma.independent
    assert sigma.get_cache_deps() == set()


def test_convert_to_parameters():
    import zfit

    conv_param1 = zfit.param.convert_to_parameters([23, 10., 34, 23], prefer_constant=True)
    assert len(conv_param1) == 4
    assert not any(p.floating for p in conv_param1)

    conv_param2 = zfit.param.convert_to_parameters([23, 10., 34, 12, 23], prefer_constant=False)
    assert all(p.floating for p in conv_param2)

    trueval3 = [23, 10., 12, 34, 23]
    truelower3 = list(range(5))
    conv_param3 = zfit.param.convert_to_parameters(trueval3,
                                                   lower=truelower3, prefer_constant=False)
    np.testing.assert_allclose(zfit.run(conv_param3), trueval3)
    np.testing.assert_allclose(zfit.run([p.lower for p in conv_param3]), truelower3)

    truename4 = ['oe', 'myname1', 'ue', 'eu', 'eue']
    stepsize4 = [23, 1.5, 10., 34, 23]
    trueupper4 = [213, 14., 1110., 314, 213]
    values4 = [23, 12, 121, 34, 23]

    conv_param4dict = zfit.param.convert_to_parameters({'value': values4,
                                                        'name': truename4,
                                                        'step_size': stepsize4,
                                                        'upper': trueupper4},
                                                       prefer_constant=False, )
    assert [p.name for p in conv_param4dict] == truename4

    np.testing.assert_allclose(zfit.run([p.upper for p in conv_param4dict]), trueupper4)
    np.testing.assert_allclose(zfit.run([p.step_size for p in conv_param4dict]), stepsize4)

    truename5 = [name + "_five" for name in truename4]
    conv_param4dict = zfit.param.convert_to_parameters(values4, name=truename5,
                                                       upper=trueupper4,
                                                       prefer_constant=False,
                                                       step_size=stepsize4)
    assert [p.name for p in conv_param4dict] == truename5

    np.testing.assert_allclose(zfit.run([p.upper for p in conv_param4dict]), trueupper4)
    np.testing.assert_allclose(zfit.run([p.step_size for p in conv_param4dict]), stepsize4)


def test_convert_to_parameters_equivalence_to_single_multi():
    import zfit

    conv_param1 = zfit.param.convert_to_parameters([23, 10., 34, 23])[1]
    assert pytest.approx(zfit.run(conv_param1.value())) == 10
    assert not conv_param1.floating

    conv_param2 = zfit.param.convert_to_parameters([23, 10., 34, 12, 23], prefer_constant=False)[-2]
    assert pytest.approx(zfit.run(conv_param2.value())) == 12.
    assert conv_param2.floating
    assert not conv_param2.has_limits

    conv_param3 = zfit.param.convert_to_parameters([23, 10., 12, 34, 23],
                                                   lower=list(range(5)), prefer_constant=False)[2]
    assert pytest.approx(zfit.run(conv_param3.lower)) == 2
    assert conv_param3.has_limits

    truename4 = ['oe', 'myname1', 'ue', 'eu', 'eue']
    stepsize4 = [23, 1.5, 10., 34, 23]
    conv_param4 = zfit.param.convert_to_parameters([23, 12, 121, 34, 23], name=truename4,
                                                   upper=[213, 14., 1110., 314, 213],
                                                   prefer_constant=False,
                                                   step_size=stepsize4)[1]
    assert conv_param4.floating
    assert conv_param4.name == 'myname1'
    assert conv_param4.has_limits
    assert conv_param4.floating
    assert pytest.approx(zfit.run(conv_param4.step_size)) == 1.5


def test_convert_to_parameters_equivalence_to_single():
    import zfit

    conv_param1 = zfit.param.convert_to_parameters(10.)[0]
    assert pytest.approx(zfit.run(conv_param1.value())) == 10
    assert not conv_param1.floating

    conv_param2 = zfit.param.convert_to_parameters(12., prefer_constant=False)[0]
    assert pytest.approx(zfit.run(conv_param2.value())) == 12.
    assert conv_param2.floating
    assert not conv_param2.has_limits

    conv_param3 = zfit.param.convert_to_parameters(12., lower=5., prefer_constant=False)[0]
    assert pytest.approx(zfit.run(conv_param3.lower)) == 5.
    assert conv_param3.has_limits

    truename4 = 'myname1'
    stepsize4 = 1.5
    conv_param4 = zfit.param.convert_to_parameters(12., name=truename4,
                                                   upper=14., prefer_constant=False, step_size=stepsize4)[0]
    assert conv_param4.floating
    assert conv_param4.name == truename4
    assert conv_param4.has_limits
    assert conv_param4.floating
    assert pytest.approx(zfit.run(conv_param4.step_size)) == stepsize4


def test_convert_to_parameter():
    import zfit

    conv_param1 = zfit.param.convert_to_parameter(10.)
    assert pytest.approx(zfit.run(conv_param1.value())) == 10
    assert not conv_param1.floating

    conv_param2 = zfit.param.convert_to_parameter(12., prefer_constant=False)
    assert pytest.approx(zfit.run(conv_param2.value())) == 12.
    assert conv_param2.floating
    assert not conv_param2.has_limits

    conv_param3 = zfit.param.convert_to_parameter(12., lower=5., prefer_constant=False)
    assert pytest.approx(zfit.run(conv_param3.lower)) == 5.
    assert conv_param3.has_limits

    with pytest.raises(ValueError):
        _ = zfit.param.convert_to_parameter(5., lower=15., prefer_constant=False)

    with pytest.raises(ValueError):
        _ = zfit.param.convert_to_parameter(5., upper=1., prefer_constant=False)

    truename4 = 'myname1'
    stepsize4 = 1.5
    conv_param4 = zfit.param.convert_to_parameter(12., name=truename4,
                                                  upper=14., prefer_constant=False, step_size=stepsize4)
    assert conv_param4.floating
    assert conv_param4.name == truename4
    assert conv_param4.has_limits
    assert conv_param4.floating
    assert pytest.approx(zfit.run(conv_param4.step_size)) == stepsize4


def test_set_values():
    import zfit
    init_values = [1, 2, 3]
    second_values = [5, 6, 7]
    params = [zfit.Parameter(f'param_{i}', val) for i, val in enumerate(init_values)]

    with zfit.param.set_values(params, second_values):
        for param, val in zip(params, second_values):
            assert param.value().numpy() == val

    for param, val in zip(params, init_values):
        assert param.value().numpy() == val

    zfit.param.set_values(params, second_values)
    for param, val in zip(params, second_values):
        assert param.value().numpy() == val

    zfit.param.set_values(params, init_values)
    for param, val in zip(params, init_values):
        assert param.value().numpy() == val



#  Copyright (c) 2021 zfit
import numpy as np
import pytest
import tensorflow as tf

from zfit import Parameter, z
from zfit.models.functions import SimpleFunc
from zfit.models.special import SimplePDF
from zfit.util.exception import BreakingAPIChangeError, ModelIncompatibleError

rnd_test_values = np.array([1., 0.01, -14.2, 0., 1.5, 152, -0.1, 12])

obs1 = 'obs1'


def test_not_allowed():
    param1 = Parameter('param1', 1.)
    param2 = Parameter('param2', 2.)
    param3 = Parameter('param3', 3., floating=False)

    def func1_pure(self, x):
        return param1 * x

    def func2_pure(self, x):
        return param2 * x + param3

    func1 = SimpleFunc(func=func1_pure, obs=obs1, p1=param1)

    pdf1 = SimplePDF(func=lambda self, x: x * param1, obs=obs1)
    pdf2 = SimplePDF(func=lambda self, x: x * param2, obs=obs1)

    with pytest.raises(BreakingAPIChangeError):
        pdf1 + pdf2

    ext_pdf1 = pdf1.create_extended(param1)
    with pytest.raises(BreakingAPIChangeError):
        ext_pdf1 + pdf2
    with pytest.raises(BreakingAPIChangeError):
        param2 * pdf1
    with pytest.raises(NotImplementedError):
        param1 + func1
    with pytest.raises(NotImplementedError):
        func1 + param1
    with pytest.raises(ModelIncompatibleError):
        func1 * pdf2
    with pytest.raises(ModelIncompatibleError):
        pdf1 * func1


def test_param_func():
    param1 = Parameter('param1', 1.)
    param2 = Parameter('param2', 2.)
    param3 = Parameter('param3', 3., floating=False)
    param4 = Parameter('param4', 4.)
    a = z.math.log(3. * param1) * tf.square(param2) - param3
    func = SimpleFunc(func=lambda self, x: a * x, obs=obs1)

    new_func = param4 * func

    new_func_equivalent = func * param4

    result1 = new_func.func(x=rnd_test_values).numpy()
    result1_equivalent = new_func_equivalent.func(x=rnd_test_values).numpy()
    result2 = func.func(x=rnd_test_values) * param4
    np.testing.assert_array_equal(result1, result2)
    np.testing.assert_array_equal(result1_equivalent, result2)


def test_func_func():
    param1 = Parameter('param1', 1.)
    param2 = Parameter('param2', 2.)
    param3 = Parameter('param3', 3., floating=False)

    def func1_pure(self, x):
        x = z.unstack_x(x)
        return param1 * x

    def func2_pure(self, x):
        x = z.unstack_x(x)
        return param2 * x + param3

    func1 = SimpleFunc(func=func1_pure, obs=obs1, p1=param1)
    func2 = SimpleFunc(func=func2_pure, obs=obs1, p2=param2, p3=param3)

    added_func = func1 + func2
    prod_func = func1 * func2

    added_values = added_func.func(rnd_test_values)
    true_added_values = func1_pure(None, rnd_test_values) + func2_pure(None, rnd_test_values)
    prod_values = prod_func.func(rnd_test_values)
    true_prod_values = func1_pure(None, rnd_test_values) * func2_pure(None, rnd_test_values)

    added_values = added_values.numpy()
    true_added_values = true_added_values.numpy()
    prod_values = prod_values.numpy()
    true_prod_values = true_prod_values.numpy()
    np.testing.assert_allclose(true_added_values, added_values)
    np.testing.assert_allclose(true_prod_values, prod_values)
